from enum import IntEnum

from blatann.nrf.nrf_types import *
from blatann.nrf.nrf_dll_load import driver
import blatann.nrf.nrf_driver_types as util
from blatann.nrf.nrf_events.gap_events import GapEvt


class GapEvtSec(GapEvt):
    pass


class GapEvtConnSecUpdate(GapEvtSec):
    evt_id = driver.BLE_GAP_EVT_CONN_SEC_UPDATE

    def __init__(self, conn_handle, sec_mode, sec_level, encr_key_size):
        super(GapEvtConnSecUpdate, self).__init__(conn_handle)
        self.sec_mode = sec_mode
        self.sec_level = sec_level
        self.encr_key_size = encr_key_size

    @classmethod
    def from_c(cls, event):
        conn_sec = event.evt.gap_evt.params.conn_sec_update.conn_sec
        return cls(conn_handle=event.evt.gap_evt.conn_handle,
                   sec_mode=conn_sec.sec_mode.sm,
                   sec_level=conn_sec.sec_mode.lv,
                   encr_key_size=conn_sec.encr_key_size)

    def __repr__(self):
        return "{}(conn_handle={!r}, sec_mode={!r}, sec_level={!r}, encr_key_size={!r})".format(self.__class__.__name__,
                                                                                                self.conn_handle,
                                                                                                self.sec_mode,
                                                                                                self.sec_level,
                                                                                                self.encr_key_size)


class GapEvtSecInfoRequest(GapEvtSec):
    evt_id = driver.BLE_GAP_EVT_SEC_INFO_REQUEST

    def __init__(self, conn_handle, peer_addr, master_id, enc_info, id_info, sign_info):
        super(GapEvtSecInfoRequest, self).__init__(conn_handle)
        self.peer_addr = peer_addr
        self.master_id = master_id
        self.enc_info = enc_info
        self.id_info = id_info
        self.sign_info = sign_info

    @classmethod
    def from_c(cls, event):
        sec_info = event.evt.gap_evt.params.sec_info_request
        conn_handle = event.evt.gap_evt.conn_handle
        peer_addr = BLEGapAddr.from_c(sec_info.peer_addr)
        master_id = BLEGapMasterId.from_c(sec_info.master_id)

        return cls(conn_handle, peer_addr, master_id, sec_info.enc_info, sec_info.id_info, sec_info.sign_info)

    def __repr__(self):
        return "{}(conn_handle={!r}, peer_addr={!r}, master_id={!r}, enc: {!r}, id: {!r}, sign: {!r})".format(
            self.__class__.__name__, self.conn_handle, self.peer_addr, self.master_id,
            self.enc_info, self.id_info, self.sign_info)


class GapEvtSecParamsRequest(GapEvtSec):
    evt_id = driver.BLE_GAP_EVT_SEC_PARAMS_REQUEST

    def __init__(self, conn_handle, sec_params):
        super(GapEvtSecParamsRequest, self).__init__(conn_handle)
        self.sec_params = sec_params

    @classmethod
    def from_c(cls, event):
        sec_params = event.evt.gap_evt.params.sec_params_request.peer_params
        return cls(conn_handle=event.evt.gap_evt.conn_handle,
                   sec_params=BLEGapSecParams.from_c(sec_params))

    def __repr__(self):
        return "{}(conn_handle={!r}, sec_params={!r})".format(self.__class__.__name__, self.conn_handle,
                                                              self.sec_params)


class GapEvtAuthKeyRequest(GapEvtSec):
    evt_id = driver.BLE_GAP_EVT_AUTH_KEY_REQUEST

    def __init__(self, conn_handle, key_type):
        super(GapEvtAuthKeyRequest, self).__init__(conn_handle)
        self.key_type = key_type

    @classmethod
    def from_c(cls, event):
        auth_key_request = event.evt.gap_evt.params.auth_key_request
        return cls(conn_handle=event.evt.gap_evt.conn_handle,
                   key_type=BLEGapAuthKeyType(auth_key_request.key_type))

    def __repr__(self):
        return "{}(conn_handle={!r}, key_type={!r})".format(self.__class__.__name__, self.conn_handle, self.key_type)


class GapEvtAuthStatus(GapEvtSec):
    evt_id = driver.BLE_GAP_EVT_AUTH_STATUS

    def __init__(self, conn_handle, auth_status, error_src, bonded, sm1_levels, sm2_levels, kdist_own, kdist_peer):
        super(GapEvtAuthStatus, self).__init__(conn_handle)
        self.auth_status = auth_status
        self.error_src = error_src
        self.bonded = bonded
        self.sm1_levels = sm1_levels
        self.sm2_levels = sm2_levels
        self.kdist_own = kdist_own
        self.kdist_peer = kdist_peer

    @classmethod
    def from_c(cls, event):
        auth_status = event.evt.gap_evt.params.auth_status
        return cls(conn_handle=event.evt.gap_evt.conn_handle,
                   auth_status=BLEGapSecStatus(auth_status.auth_status),
                   error_src=auth_status.error_src,
                   bonded=auth_status.bonded,
                   sm1_levels=BLEGapSecLevels.from_c(auth_status.sm1_levels),
                   sm2_levels=BLEGapSecLevels.from_c(auth_status.sm2_levels),
                   kdist_own=BLEGapSecKeyDist.from_c(auth_status.kdist_own),
                   kdist_peer=BLEGapSecKeyDist.from_c(auth_status.kdist_peer))

    def __repr__(self):
        return "{}(conn_handle={!r}, auth_status={!r}, error_src={!r}, " \
               "bonded={!r}, sm1_levels={!r}, sm2_levels={!r}, " \
               "kdist_own={!r}, kdist_peer={!r})".format(self.__class__.__name__, self.conn_handle, self.auth_status,
                                                         self.error_src, self.bonded,
                                                         self.sm1_levels, self.sm2_levels, self.kdist_own,
                                                         self.kdist_peer)


class GapEvtPasskeyDisplay(GapEvtSec):
    evt_id = driver.BLE_GAP_EVT_PASSKEY_DISPLAY

    def __init__(self, conn_handle, passkey, match_request):
        super(GapEvtPasskeyDisplay, self).__init__(conn_handle)
        self.passkey = passkey
        self.match_request = match_request

    @classmethod
    def from_c(cls, event):
        passkey_display = event.evt.gap_evt.params.passkey_display
        passkey = "".join(chr(c) for c in util.uint8_array_to_list(passkey_display.passkey, 6))
        match_request = bool(passkey_display.match_request)
        return cls(conn_handle=event.evt.gap_evt.conn_handle,
                   passkey=passkey,
                   match_request=match_request)

    def __repr__(self):
        return "{}(conn_handle={!r}, passkey={!r}, match_request={!r})".format(self.__class__.__name__,
                                                                               self.conn_handle, self.passkey,
                                                                               self.match_request)


class GapEvtLescDhKeyRequest(GapEvtSec):
    evt_id = driver.BLE_GAP_EVT_LESC_DHKEY_REQUEST

    def __init__(self, conn_handle, remote_public_key, oob_required):
        super(GapEvtLescDhKeyRequest, self).__init__(conn_handle)
        self.remote_public_key = remote_public_key  # type: BLEGapDhKey
        self.oob_required = oob_required

    @classmethod
    def from_c(cls, event):
        dh_key_request = event.evt.gap_evt.params.lesc_dhkey_request
        remote_pk = BLEGapPublicKey.from_c(dh_key_request.p_pk_peer)
        oob_required = bool(dh_key_request.oobd_req)
        return cls(conn_handle=event.evt.gap_evt.conn_handle,
                   remote_public_key=remote_pk,
                   oob_required=oob_required)

    def __repr__(self):
        return "{}(conn_handle={!r}, remote key: {!r}, oob? {}".format(self.__class__.__name__, self.conn_handle,
                                                                       self.remote_public_key, self.oob_required)



from blatann.services.device_info.service import DisClient as _DisClient, DisServer as _DisServer
from blatann.services.device_info.constants import (
    DIS_SERVICE_UUID, CHARACTERISTICS,
    SystemIdCharacteristic, ModelNumberCharacteristic, SerialNumberCharacteristic, FirmwareRevisionCharacteristic,
    HardwareRevisionCharacteristic, SoftwareRevisionCharacteristic, ManufacturerNameCharacteristic,
    RegulatoryCertificateCharacteristic, PnpIdCharacteristic
)
from blatann.services.device_info.data_types import PnpId, SystemId, PnpVendorSource


def add_device_info_service(gatts_database):
    """
    :rtype: _DisServer
    """
    return _DisServer.add_to_database(gatts_database)


def find_device_info_service(gattc_database):
    """
    :type gattc_database: blatann.gatt.gattc.GattcDatabase
    :rtype: _DisClient
    """
    return _DisClient.find_in_database(gattc_database)



import threading


def _or(self, other):
    if isinstance(self, _OrEvent):
        events = self.events[:]
    elif isinstance(self, threading.Event):
        events = [self]
    else:
        raise ValueError("Incompatible Event type to OR with")

    if isinstance(other, _OrEvent):
        events.extend(other.events)
    elif isinstance(other, threading.Event):
        events.append(other)
    else:
        raise ValueError("Incompatible Event type to OR with")

    return _OrEvent(*events)


def _or_set(self):
    self._set()
    for h in self._changed:
        h()


def _or_clear(self):
    self._clear()
    for h in self._changed:
        h()


class _OrEvent(threading.Event):
    def __init__(self, *events):
        super(_OrEvent, self).__init__()
        self.events = []
        for e in events:
            self.add(e)

    def _changed(self):
        bools = [e.is_set() for e in self.events]
        if any(bools):
            self.set()
        else:
            self.clear()

    def _orify(self, e):
        if not hasattr(e, "_set"):
            e._set = e.set
            e._clear = e.clear
            e._changed = []
            e.set = lambda: _or_set(e)
            e.clear = lambda: _or_clear(e)
        e._changed.append(self._changed)

    def add(self, event):
        if isinstance(event, _OrEvent):
            for e in event.events:
                self._orify(e)
            self.events.extend(event.events)
        else:
            self._orify(event)
            self.events.append(event)

    def __or__(self, other):
        return _or(self, other)


threading.Event.__or__ = _or



import enum
import logging
import struct
from blatann.nrf.nrf_types.gatt import BLE_GATT_HANDLE_INVALID
from blatann.nrf import nrf_types
from blatann.gap.smp import SecurityLevel


logger = logging.getLogger(__name__)

"""
The default MTU size that's used when a connection is established
"""
MTU_SIZE_DEFAULT = 23

"""
The minimum allowed MTU size
"""
MTU_SIZE_MINIMUM = 23

"""
The ideal MTU size to use when using the maximum link-layer Data Length Extension setting (251) 
"""
MTU_SIZE_FOR_MAX_DLE = 247

# Overhead counts for different ATT processes
WRITE_BYTE_OVERHEAD = 3
LONG_WRITE_BYTE_OVERHEAD = 5
NOTIFICATION_BYTE_OVERHEAD = 3
READ_BYTE_OVERHEAD = 1

"""
Status codes that can be returned during GATT Operations (reads, writes, etc.)
"""
GattStatusCode = nrf_types.BLEGattStatusCode

"""
The two notification types (notification, indication) used when a characteristic is notified from a peripheral
"""
GattNotificationType = nrf_types.BLEGattHVXType


class ServiceType(enum.Enum):
    PRIMARY = 1
    SECONDARY = 2


class SubscriptionState(enum.IntEnum):
    """
    Defines the different subscription states/types for a characteristic
    """
    NOT_SUBSCRIBED = 0
    NOTIFY = 1
    INDICATION = 2

    @classmethod
    def to_buffer(cls, value):
        """
        Converts to a little-endian uint16 buffer to be written over BLE
        """
        return struct.pack("<H", value)

    @classmethod
    def from_buffer(cls, buf):
        """
        Converts from a little-endian uint16 buffer received over BLE to the subscription state
        """
        return cls(struct.unpack("<H", buf)[0])


class CharacteristicProperties(object):
    def __init__(self, read=True, write=False, notify=False, indicate=False, broadcast=False,
                 write_no_response=False, signed_write=False):
        self.read = read
        self.write = write
        self.notify = notify
        self.indicate = indicate
        self.broadcast = broadcast
        self.write_no_response = write_no_response
        self.signed_write = signed_write

    @classmethod
    def from_nrf_properties(cls, nrf_props):
        """
        :type nrf_props: blatann.nrf.nrf_types.BLEGattCharacteristicProperties
        """
        return CharacteristicProperties(nrf_props.read, nrf_props.write, nrf_props.notify, nrf_props.indicate,
                                        nrf_props.broadcast, nrf_props.write_wo_resp, nrf_props.auth_signed_wr)

    def __repr__(self):
        props = [
            [self.read, "r"],
            [self.write, "w"],
            [self.notify, "n"],
            [self.indicate, "i"],
            [self.broadcast, "b"],
            [self.write_no_response, "wn"],
            [self.signed_write, "sw"],
        ]
        props = [c for is_set, c in props if is_set]
        return "CharProps({})".format(",".join(props))


class Characteristic(object):
    """
    Abstract class that represents a BLE characteristic (both remote and local)
    """
    def __init__(self, ble_device, peer, uuid, properties,
                 default_string_encoding="utf8"):
        """
        :type ble_device: blatann.device.BleDevice
        :type peer: blatann.peer.Peer
        :type uuid: blatann.uuid.Uuid
        :type properties: CharacteristicProperties
        """
        self.ble_device = ble_device
        self.peer = peer
        self.uuid = uuid
        self.declaration_handle = BLE_GATT_HANDLE_INVALID
        self.value_handle = BLE_GATT_HANDLE_INVALID
        self.cccd_handle = BLE_GATT_HANDLE_INVALID
        self.cccd_state = SubscriptionState.NOT_SUBSCRIBED
        self._properties = properties
        self._string_encoding = default_string_encoding

    @property
    def string_encoding(self):
        """
        The default method for encoding strings into bytes when a string is provided as a value
        """
        return self._string_encoding

    @string_encoding.setter
    def string_encoding(self, encoding):
        """
        Sets how the characteristic value is encoded when provided a string

        :param encoding: the encoding to use (utf8, ascii, etc.)
        """
        self._string_encoding = encoding

    def __repr__(self):
        return "Characteristic({}, {}".format(self.uuid, self._properties)


class Service(object):
    """
    Abstract class that represents a BLE Service (both remote and local)
    """
    def __init__(self, ble_device, peer, uuid, service_type,
                 start_handle=BLE_GATT_HANDLE_INVALID, end_handle=BLE_GATT_HANDLE_INVALID):
        """
        :type ble_device: blatann.device.BleDevice
        :type peer: blatann.peer.Peer
        """
        self.ble_device = ble_device
        self.peer = peer
        self.uuid = uuid
        self.service_type = service_type
        self._characteristics = []
        self.start_handle = start_handle
        # If a valid starting handle is given and not a valid ending handle, then the ending handle
        # is the starting handle
        if start_handle != BLE_GATT_HANDLE_INVALID and end_handle == BLE_GATT_HANDLE_INVALID:
            end_handle = start_handle
        self.end_handle = end_handle

    def __repr__(self):
        return "Service({}, characteristics: [{}])".format(self.uuid,
                                                           "\n    ".join(str(c) for c in self._characteristics))


class GattDatabase(object):
    """
    Abstract class that represents a BLE Database (both remote and local)
    """
    def __init__(self, ble_device, peer):
        """
        :type ble_device: blatann.device.BleDevice
        :type peer: blatann.peer.Peer
        """
        self.ble_device = ble_device
        self.peer = peer
        self._services = []

    def __repr__(self):
        return "Database(peer {}, services: [{}])".format(self.peer.conn_handle,
                                                          "\n  ".join(str(s) for s in self._services))




from blatann.nrf.nrf_events.generic_events import *
from blatann.nrf.nrf_events.gap_events import *
from blatann.nrf.nrf_events.smp_events import *
from blatann.nrf.nrf_events.gatt_events import *


_event_classes = [
    EvtUserMemoryRequest,

    # Gap
    GapEvtConnected,
    GapEvtDisconnected,
    GapEvtConnParamUpdate,
    GapEvtAdvReport,
    GapEvtTimeout,
    GapEvtConnParamUpdateRequest,
    GapEvtDataLengthUpdate,
    GapEvtDataLengthUpdateRequest,
    GapEvtPhyUpdate,
    GapEvtPhyUpdateRequest,

    # SMP
    GapEvtSecParamsRequest,
    GapEvtAuthKeyRequest,
    GapEvtConnSecUpdate,
    GapEvtAuthStatus,
    GapEvtPasskeyDisplay,
    GapEvtSecInfoRequest,
    GapEvtLescDhKeyRequest,
    # TODO: (not all listed)
    # driver.BLE_GAP_EVT_SEC_REQUEST,

    # Gattc
    GattcEvtPrimaryServiceDiscoveryResponse,
    GattcEvtCharacteristicDiscoveryResponse,
    GattcEvtDescriptorDiscoveryResponse,
    GattcEvtReadResponse,
    GattcEvtWriteResponse,
    GattcEvtHvx,
    GattcEvtAttrInfoDiscoveryResponse,
    GattcEvtMtuExchangeResponse,
    GattcEvtTimeout,
    # TODO:
    # driver.BLE_GATTC_EVT_REL_DISC_RSP
    # driver.BLE_GATTC_EVT_CHAR_VAL_BY_UUID_READ_RSP
    # driver.BLE_GATTC_EVT_CHAR_VALS_READ_RSP
    # driver.BLE_GATTC_EVT_WRITE_CMD_TX_COMPLETE

    # Gatts
    GattsEvtWrite,
    GattsEvtReadWriteAuthorizeRequest,
    GattsEvtHandleValueConfirm,
    GattsEvtExchangeMtuRequest,
    GattsEvtNotificationTxComplete,
    GattsEvtTimeout,
    # TODO:
    # driver.BLE_GATTS_SYS_ATTR_MISSING
    # driver.BLE_GATTS_EVT_SC_CONFIRM
]

_events_by_id = {e.evt_id: e for e in _event_classes}


def event_decode(ble_event):
    event_cls = _events_by_id.get(ble_event.header.evt_id, None)
    if event_cls:
        return event_cls.from_c(ble_event)
    return None



import logging
from blatann.gatt import MTU_SIZE_DEFAULT, WRITE_BYTE_OVERHEAD
from blatann.gatt.gatts import GattsService, GattsCharacteristicProperties
from blatann.event_type import EventSource, Event
from blatann import exceptions
from blatann.services.ble_data_types import Uint16, BleDataStream
from blatann.services.nordic_uart.constants import *
from blatann.waitables.event_waitable import EventWaitable


logger = logging.getLogger(__name__)


class NordicUartServer(object):
    def __init__(self, service, max_characteristic_size=None):
        """
        :type service: GattsService
        """
        self._service = service
        if max_characteristic_size is None:
            max_characteristic_size = self._service.peer.max_mtu_size - WRITE_BYTE_OVERHEAD

        rx_props = GattsCharacteristicProperties(read=False, write=False, notify=True, indicate=False,
                                                 variable_length=True, max_length=max_characteristic_size)

        self._rx_char = service.add_characteristic(NORDIC_UART_RX_CHARACTERISTIC_UUID, rx_props)

        tx_props = GattsCharacteristicProperties(read=False, write=True, notify=False, indicate=False,
                                                 variable_length=True, max_length=max_characteristic_size)

        self._tx_char = service.add_characteristic(NORDIC_UART_TX_CHARACTERISTIC_UUID, tx_props)

        feature_props = GattsCharacteristicProperties(read=True, write=False, notify=False, indicate=False,
                                                      variable_length=False, max_length=Uint16.byte_count)
        self._feature_char = service.add_characteristic(NORDIC_UART_FEATURE_CHARACTERISTIC_UUID, feature_props,
                                                        Uint16.encode(max_characteristic_size))

        self._on_data_received = EventSource("Data Received", logger)
        self._on_write_complete = EventSource("Write Complete", logger)
        self._tx_char.on_write.register(self._on_write)

    @property
    def on_data_received(self):
        """
        :rtype: Event
        """
        return self._on_data_received

    @property
    def on_write_complete(self):
        """
        :rtype: Event
        """
        return self._rx_char.on_notify_complete

    @property
    def max_write_length(self):
        return self._rx_char.max_length

    def write(self, data):
        if len(data) > self.max_write_length:
            raise ValueError("Can only write max {} bytes at a time".format(self.max_write_length))
        return self._rx_char.notify(data)

    def _on_write(self, characteristic, event_args):
        self._on_data_received.notify(self, event_args.value)

    @classmethod
    def add_to_database(cls, gatts_database, max_characteristic_size=None):
        service = gatts_database.add_service(NORDIC_UART_SERVICE_UUID)
        return cls(service, max_characteristic_size)


class NordicUartClient(object):
    def __init__(self, service):
        """
        :type service: blatann.gatt.gattc.GattcService
        """
        self._service = service
        self._server_characteristic_size = MTU_SIZE_DEFAULT
        self._tx_char = self._service.find_characteristic(NORDIC_UART_TX_CHARACTERISTIC_UUID)
        self._rx_char = self._service.find_characteristic(NORDIC_UART_RX_CHARACTERISTIC_UUID)
        self._feature_char = self._service.find_characteristic(NORDIC_UART_FEATURE_CHARACTERISTIC_UUID)
        self._init_complete_event = EventSource("Init Complete", logger)
        self._on_data_received = EventSource("Data Received", logger)
        self._on_write_complete = EventSource("Write Complete", logger)

    @property
    def on_data_received(self):
        """
        :rtype: Event
        """
        return self._on_data_received

    @property
    def on_write_complete(self):
        """
        :rtype: Event
        """
        return self._tx_char.on_write_complete

    @property
    def max_write_length(self):
        return self._server_characteristic_size

    @property
    def is_initialized(self):
        return self._rx_char.subscribed

    def initialize(self):
        self._rx_char.subscribe(self._on_notify_received).then(self._check_feature_characteristic)
        return EventWaitable(self._init_complete_event)

    def write(self, data):
        if not self.is_initialized:
            raise exceptions.InvalidStateException("Service must be initialized before write")
        if len(data) > self.max_write_length:
            raise ValueError("Can only write max {} bytes at a time".format(self.max_write_length))
        return self._tx_char.write(data)

    def _check_feature_characteristic(self, characteristic, event_args):
        if not self._feature_char:
            self._init_complete_event.notify(self, event_args.value)
        else:
            self._feature_char.read().then(self._process_feature_value)

    def _process_feature_value(self, characteristic, event_args):
        self._server_characteristic_size = Uint16.decode(BleDataStream(event_args.value))
        self._init_complete_event.notify(self, None)

    def _on_notify_received(self, characteristic, event_args):
        self._on_data_received.notify(self, event_args.value)

    @classmethod
    def find_in_database(cls, gattc_database):
        """
        :type gattc_database: blatann.gatt.gattc.GattcDatabase
        :rtype: NordicUartClient
        """
        service = gattc_database.find_service(NORDIC_UART_SERVICE_UUID)
        if service:
            return cls(service)



from blatann.utils import setup_logger


def find_target_device(ble_device, name):
    """
    Starts the scanner and searches the advertising report for the desired name.
    If found, returns the peer's address that can be connected to

    :param ble_device: The ble device to operate on
    :type ble_device: blatann.BleDevice
    :param name: The device's local name that is advertised
    :return: The peer's address if found, or None if not found
    """
    scan_report = ble_device.scanner.start_scan().wait()

    for report in scan_report.advertising_peers_found:
        if report.advertise_data.local_name == name:
            return report.peer_address



from typing import Iterable, List
import logging
from blatann.nrf import nrf_types
from blatann import uuid, exceptions


logger = logging.getLogger(__name__)


class AdvertisingFlags(object):
    LIMITED_DISCOVERY_MODE = 0x01
    GENERAL_DISCOVERY_MODE = 0x02
    BR_EDR_NOT_SUPPORTED = 0x04
    BR_EDR_CONTROLLER = 0x08
    BR_EDR_HOST = 0x10


class AdvertisingData(object):
    """
    Class which represents data that can be advertised
    """
    MAX_ENCODED_LENGTH = 31  # Bluetooth-defined max length that the encoded data can be

    Types = nrf_types.BLEAdvData.Types  # Enum representing the different advertising data types

    def __init__(self, flags=None, local_name=None, local_name_complete=True,
                 service_uuid16s=None, service_uuid128s=None,
                 has_more_uuid16_services=False, has_more_uuid128_services=False,
                 service_data=None, manufacturer_data=None, **other_flags):
        self.flags = flags
        self.local_name = local_name
        self.local_name_complete = local_name_complete
        self.service_uuid16s = service_uuid16s or []
        self.service_uuid128s = service_uuid128s or []
        self.has_more_uuid16_services = has_more_uuid16_services
        self.has_more_uuid128_services = has_more_uuid128_services
        self.service_data = service_data
        self.manufacturer_data = manufacturer_data
        self.other_flags = {self.Types[k]: v for k, v in other_flags.items()}
        if not isinstance(self.service_uuid16s, (list, tuple)):
            self.service_uuid16s = [self.service_uuid16s]
        if not isinstance(self.service_uuid128s, (list, tuple)):
            self.service_uuid128s = [self.service_uuid128s]

    @property
    def service_uuids(self):
        """
        Gets all of the 16-bit and 128-bit service UUIDs specified in the advertising data

        :return: list of the service UUIDs present in the advertising data
        :rtype: list[uuid.Uuid]
        """
        return self.service_uuid16s + self.service_uuid128s

    def check_encoded_length(self):
        """
        Checks if the encoded length of this advertising data payload meets the maximum allowed
        length specified by the Bluetooth spec

        :return: a tuple of the encoded length and a bool result of whether or not it meets requirements
        :rtype: tuple[int, bool]
        """
        ble_adv_data = self.to_ble_adv_data()
        encoded_data = ble_adv_data.to_list()
        return len(encoded_data), len(encoded_data) <= self.MAX_ENCODED_LENGTH

    def to_ble_adv_data(self):
        """
        Converts the advertising data to a BLEAdvData object that can be used by the nRF Driver

        :return: the BLEAdvData object which represents this advertising data
        :rtype: nrf_types.BLEAdvData
        """
        records = self.other_flags.copy()
        if self.flags:
            records[self.Types.flags] = [int(self.flags)]
        if self.local_name:
            name_type = self.Types.complete_local_name if self.local_name_complete else self.Types.short_local_name
            records[name_type] = self.local_name
        if self.service_uuid128s:
            uuid_type = self.Types.service_128bit_uuid_more_available if self.has_more_uuid128_services else self.Types.service_128bit_uuid_complete
            # UUID data is little-endian, reverse the lists and concatenate
            uuid_data = []
            for u in self.service_uuid128s:
                uuid_data.extend(u.uuid[::-1])
            records[uuid_type] = uuid_data
        if self.service_uuid16s:
            uuid_type = self.Types.service_16bit_uuid_more_available if self.has_more_uuid16_services else self.Types.service_16bit_uuid_complete
            uuid_data = []
            for u in self.service_uuid16s:
                uuid_data.append(u.uuid & 0xFF)
                uuid_data.append((u.uuid >> 8) & 0xFF)
            records[uuid_type] = uuid_data
        if self.service_data:
            records[self.Types.service_data] = self.service_data
        if self.manufacturer_data:
            records[self.Types.manufacturer_specific_data] = self.manufacturer_data

        record_string_keys = {k.name: v for k, v in records.items()}
        return nrf_types.BLEAdvData(**record_string_keys)

    @classmethod
    def from_ble_adv_records(cls, advertise_records):
        """
        Converts a dictionary of AdvertisingData.Type: value keypairs into an object of this class

        :param advertise_records: a dictionary mapping the advertise data types to their corresponding values
        :type advertise_records: dict
        :return: the AdvertisingData from the records given
        :rtype: AdvertisingData
        """
        flags = advertise_records.pop(cls.Types.flags, None)
        if flags:
            flags = flags[0]

        local_name_complete = False
        local_name = advertise_records.pop(cls.Types.complete_local_name, None)
        if local_name:
            local_name_complete = True
        else:
            local_name = advertise_records.pop(cls.Types.short_local_name, None)
        if local_name:
            local_name = str("".join(chr(c) for c in local_name))

        manufacturer_data = advertise_records.pop(cls.Types.manufacturer_specific_data, None)
        if manufacturer_data:
            manufacturer_data = bytearray(manufacturer_data)

        service_data = advertise_records.pop(cls.Types.service_data, None)
        if service_data:
            service_data = bytearray(service_data)

        more_16bit_services = False
        uuid16_data = advertise_records.pop(cls.Types.service_16bit_uuid_more_available, None)
        if uuid16_data:
            more_16bit_services = True
        else:
            uuid16_data = advertise_records.pop(cls.Types.service_16bit_uuid_complete, None)

        service_uuid16s = []
        if uuid16_data:
            for i in range(0, len(uuid16_data), 2):
                uuid16 = (uuid16_data[i+1] << 8) | uuid16_data[i]
                service_uuid16s.append(uuid.Uuid16(uuid16))

        more_128bit_services = False
        uuid128_data = advertise_records.pop(cls.Types.service_128bit_uuid_more_available, None)
        if uuid128_data:
            more_128bit_services = True
        else:
            uuid128_data = advertise_records.pop(cls.Types.service_128bit_uuid_complete, None)

        service_uuid128s = []
        if uuid128_data:
            for i in range(0, len(uuid128_data), 16):
                uuid128 = uuid128_data[i:i+16][::-1]
                service_uuid128s.append(uuid.Uuid128(uuid128))

        record_string_keys = {k.name: bytearray(v) for k, v in advertise_records.items()}
        return AdvertisingData(flags=flags, local_name=local_name, local_name_complete=local_name_complete,
                               service_uuid16s=service_uuid16s, service_uuid128s=service_uuid128s,
                               has_more_uuid16_services=more_16bit_services, has_more_uuid128_services=more_128bit_services,
                               service_data=service_data, manufacturer_data=manufacturer_data, **record_string_keys)

    def __repr__(self):
        params = []
        if self.local_name:
            name_str = "Name: {}".format(self.local_name)
            if not self.local_name_complete:
                name_str += "(short)"
            params.append(name_str)
        if self.flags:
            params.append("Flags: {}".format(self.flags))
        if self.service_uuid16s:
            param_str = "16-bit Services: [{}]".format(", ".join(str(u) for u in self.service_uuid16s))
            if self.has_more_uuid16_services:
                param_str += "+ more"
            params.append(param_str)
        if self.service_uuid128s:
            param_str = "128-bit Services: [{}]".format(", ".join(str(u) for u in self.service_uuid128s))
            if self.has_more_uuid128_services:
                param_str += "+ more"
            params.append(param_str)
        if self.service_data:
            params.append("Service Data: {}".format(self.service_data))

        return "{}({})".format(self.__class__.__name__, ", ".join(params))


class ScanReport(object):
    def __init__(self, adv_report):
        """
        :type adv_report: blatann.nrf.nrf_events.GapEvtAdvReport
        """
        self.peer_address = adv_report.peer_addr
        self._current_advertise_data = adv_report.adv_data.records.copy()
        self.advertise_data = AdvertisingData.from_ble_adv_records(self._current_advertise_data)
        self.rssi = adv_report.rssi
        self.duplicate = False

    @property
    def device_name(self):
        return self.advertise_data.local_name or str(self.peer_address)

    def update(self, adv_report):
        """
        :type adv_report: nrf_events.GapEvtAdvReport
        """
        if adv_report.peer_addr != self.peer_address:
            raise exceptions.InvalidOperationException("Peer address doesn't match")
        self._current_advertise_data.update(adv_report.adv_data.records)
        self.advertise_data = AdvertisingData.from_ble_adv_records(self._current_advertise_data.copy())
        self.rssi = max(self.rssi, adv_report.rssi)

    def __eq__(self, other):
        return self.peer_address == other.peer_address and self._current_advertise_data == other._current_advertise_data

    def __repr__(self):
        return "{}: {}dBm - {}".format(self.device_name, self.rssi, self.advertise_data)


class ScanReportCollection(object):
    """
    Collection of all the advertising data and scan reports found in a scanning session
    """
    def __init__(self):
        self._all_scans = []
        self._scans_by_peer_address = {}

    @property
    def advertising_peers_found(self) -> Iterable[ScanReport]:
        """
        Gets the list of scans which have been combined and condensed into a list where each entry is a unique peer

        :return: The list of scan reports, with each being a unique peer
        :rtype: list of ScanReport
        """
        return self._scans_by_peer_address.values()

    @property
    def all_scan_reports(self) -> List[ScanReport]:
        """
        Gets the list of all scanned advertising data found.

        :return: The list of all scan reports
        :rtype: list of ScanReport
        """
        return self._all_scans

    def clear(self):
        """
        Clears out all of the scan reports cached
        """
        self._all_scans = []
        self._scans_by_peer_address = {}

    def update(self, adv_report):
        """
        :type adv_report: nrf_events.GapEvtAdvReport
        """
        scan_entry = ScanReport(adv_report)
        if scan_entry in self._all_scans:
            scan_entry.duplicate = True
        self._all_scans.append(scan_entry)
        if adv_report.peer_addr in self._scans_by_peer_address.keys():
            self._scans_by_peer_address[adv_report.peer_addr].update(adv_report)
        else:
            self._scans_by_peer_address[adv_report.peer_addr] = scan_entry
        return self._scans_by_peer_address[adv_report.peer_addr]



from enum import IntEnum
from blatann.services import ble_data_types
from blatann.exceptions import DecodeError


class RacpOpcode(IntEnum):
    report_stored_records = 1
    delete_stored_records = 2
    abort_operation = 3
    report_number_of_records = 4
    number_of_records_response = 5
    response_code = 6


class RacpOperator(IntEnum):
    null = 0
    all_records = 1
    less_than_or_equal_to = 2
    greater_than_or_equal_to = 3
    within_range_inclusive = 4
    first_record = 5
    last_record = 6


class FilterType(IntEnum):
    sequence_number = 1
    user_facing_time = 2


class RacpResponseCode(IntEnum):
    success = 1
    not_supported = 2
    invalid_operator = 3
    operator_not_supported = 4
    invalid_operand = 5
    no_records_found = 6
    abort_not_successful = 7
    procedure_not_completed = 8
    operand_not_supported = 9


class RacpCommand(ble_data_types.BleCompoundDataType):
    def __init__(self, opcode, operator, filter_type=None, filter_params=None):
        """
        :type opcode: RacpOpcode
        :type operator: RacpOperator
        :type filter_type: FilterType
        :param filter_params:
        """
        self.opcode = opcode
        self.operator = operator
        self.filter_type = filter_type
        if filter_params is None:
            filter_params = []
        self.filter_params = filter_params

    def get_filter_min_max(self):
        if self.operator == RacpOperator.all_records:
            return None, None
        if self.operator == RacpOperator.less_than_or_equal_to:
            return None, self.filter_params[0]
        if self.operator == RacpOperator.greater_than_or_equal_to:
            return self.filter_params[0], None
        if self.operator == RacpOperator.within_range_inclusive:
            return self.filter_params[0], self.filter_params[1]
        # First/Last record, return Nones
        return None, None

    def encode(self):
        stream = ble_data_types.BleDataStream()
        stream.encode(ble_data_types.Uint8, self.opcode)
        stream.encode(ble_data_types.Uint8, self.operator)
        if self.filter_type is not None:
            stream.encode(ble_data_types.Uint8, self.filter_type)
            for f in self.filter_params:
                stream.encode(ble_data_types.Uint16, f)
        return stream

    @classmethod
    def decode(cls, stream):
        opcode = stream.decode(ble_data_types.Uint8)
        operator = stream.decode(ble_data_types.Uint8)
        if len(stream) > 0:
            filter_type = stream.decode(ble_data_types.Uint8)
            filter_params = []
            while len(stream) >= 2:
                filter_params.append(stream.decode(ble_data_types.Uint16))
        else:
            filter_type = None
            filter_params = None

        return RacpCommand(opcode, operator, filter_type, filter_params)


class RacpResponse(ble_data_types.BleCompoundDataType):
    def __init__(self, request_opcode=None, response_code=None, record_count=None):
        """
        :type request_opcode: RacpOpcode
        :type response_code: RacpResponseCode
        :type record_count: int
        """
        self.request_code = request_opcode
        self.response_code = response_code
        self.record_count = record_count

    def encode(self):
        stream = ble_data_types.BleDataStream()
        if self.record_count is None:
            stream.encode_multiple([ble_data_types.Uint8, RacpOpcode.response_code],
                                   [ble_data_types.Uint8, RacpOperator.null],
                                   [ble_data_types.Uint8, self.request_code],
                                   [ble_data_types.Uint8, self.response_code])
        else:
            stream.encode_multiple([ble_data_types.Uint8, RacpOpcode.number_of_records_response],
                                   [ble_data_types.Uint8, RacpOperator.null],
                                   [ble_data_types.Uint16, self.record_count])
        return stream

    @classmethod
    def decode(cls, stream):
        opcode = RacpOpcode(stream.decode(ble_data_types.Uint8))
        _ = RacpOperator(stream.decode(ble_data_types.Uint8))

        if opcode == RacpOpcode.response_code:
            request_opcode, response_code = stream.decode_multiple(ble_data_types.Uint8, ble_data_types.Uint8)
            record_count = None
        elif opcode == RacpOpcode.number_of_records_response:
            request_opcode = None
            response_code = None
            record_count = stream.decode(ble_data_types.Uint16)
        else:
            raise DecodeError("Unable to decode RACP Response, got opcode: {}".format(opcode))

        return RacpResponse(request_opcode, response_code, record_count)



from enum import Enum
import logging

from pc_ble_driver_py.lib import nrf_ble_driver_sd_api_v5

from blatann.nrf import nrf_driver_types as util
from blatann.nrf.nrf_dll_load import driver

logger = logging.getLogger(__name__)


class BleOptionFlag(Enum):
    pa_lna = driver.BLE_COMMON_OPT_PA_LNA
    conn_event_extension = driver.BLE_COMMON_OPT_CONN_EVT_EXT
    gap_channel_map = driver.BLE_GAP_OPT_CH_MAP
    gap_local_conn_latency = driver.BLE_GAP_OPT_LOCAL_CONN_LATENCY
    gap_passkey = driver.BLE_GAP_OPT_PASSKEY
    gap_scan_req_report = driver.BLE_GAP_OPT_SCAN_REQ_REPORT
    gap_compat_mode_1 = driver.BLE_GAP_OPT_COMPAT_MODE_1
    gap_auth_payload_timeout = driver.BLE_GAP_OPT_AUTH_PAYLOAD_TIMEOUT
    gap_slave_latency_disable = driver.BLE_GAP_OPT_SLAVE_LATENCY_DISABLE


class BleOption(object):
    option_flag = None

    def get_opt_subfield(self, opt: driver.ble_opt_t):
        raise NotImplementedError

    def to_c(self):
        raise NotImplementedError


class BleEnableOpt(BleOption):
    _default = False
    _driver_type = None

    def __init__(self, enabled=_default):
        self.enabled = enabled

    def get_opt_subfield(self, opt: driver.ble_opt_t):
        raise NotImplementedError

    def to_c(self):
        opt = self._driver_type()
        opt.enable = self.enabled
        return opt


class BleOptConnEventExtenion(BleEnableOpt):
    option_flag = BleOptionFlag.conn_event_extension
    _driver_type = driver.ble_common_opt_conn_evt_ext_t

    def get_opt_subfield(self, opt: driver.ble_opt_t):
        return opt.common_opt.conn_evt_ext


class BlePaLnaConfig(object):
    def __init__(self, enabled=False, active_high=True, pin=0):
        self.enabled = enabled
        self.active_high = active_high
        self.pin = pin

    def to_c(self):
        cfg = driver.ble_pa_lna_cfg_t()
        cfg.enable = int(self.enabled)
        cfg.active_high = int(self.active_high)
        cfg.gpio_pin = self.pin
        return cfg


class BleOptPaLna(BleOption):
    option_flag = BleOptionFlag.pa_lna

    def __init__(self, pa_config=None, lna_cfg=None, ppi_channel_set=0,
                 ppi_channel_clear=0, gpiote_channel=0):
        self.pa_config = pa_config or BlePaLnaConfig()
        self.lna_config = lna_cfg or BlePaLnaConfig()
        self.ppi_channel_set = ppi_channel_set
        self.ppi_channel_clear = ppi_channel_clear
        self.gpiote_channel = gpiote_channel

    def get_opt_subfield(self, opt: driver.ble_opt_t):
        return opt.common_opt.pa_lna

    def to_c(self):
        opt = driver.ble_common_opt_pa_lna_t()
        opt.pa_cfg = self.pa_config.to_c()
        opt.lna_cfg = self.lna_config.to_c()
        opt.ppi_ch_id_set = self.ppi_channel_set
        opt.ppi_ch_id_clr = self.ppi_channel_clear
        opt.gpiote_ch_id = self.gpiote_channel
        return opt


class BleOptGapChannelMap(BleOption):
    option_flag = BleOptionFlag.gap_channel_map

    def __init__(self, enabled_channels=None, conn_handle=0):
        self.conn_handle = conn_handle
        self.channel_map = enabled_channels or list(range(37))

    def get_opt_subfield(self, opt: driver.ble_opt_t):
        return opt.gap_opt.ch_map

    def to_c(self):
        opt = driver.ble_gap_opt_ch_map_t()
        opt.conn_handle = self.conn_handle
        for i in self.channel_map:
            if i >= 37:
                logger.warning("Cannot set channel {} in the channel map".format(i))
                continue
            byte = i // 8
            bit = 1 << (i % 8)
            opt.ch_map[byte] |= bit

        return opt


class BleOptGapLocalConnLatency(BleOption):
    option_flag = BleOptionFlag.gap_local_conn_latency

    def __init__(self, conn_handle=0, requested_latency=0):
        self.conn_handle = conn_handle
        self.requested_latency = requested_latency

    def get_opt_subfield(self, opt: driver.ble_opt_t):
        return opt.gap_opt.local_conn_latency

    def to_c(self):
        opt = driver.ble_gap_opt_local_conn_latency_t()
        opt.conn_handle = self.conn_handle
        opt.requested_latency = self.requested_latency
        return opt


class BleOptGapPasskey(BleOption):
    option_flag = BleOptionFlag.gap_passkey

    def __init__(self, passkey="000000"):
        self.passkey = passkey

    def get_opt_subfield(self, opt: driver.ble_opt_t):
        return opt.gap_opt.passkey

    def to_c(self):
        opt = driver.ble_gap_opt_passkey_t()
        opt.p_passkey = util.list_to_char_array(self.passkey).cast()
        return opt


class BleOptGapScanRequestReport(BleEnableOpt):
    option_flag = BleOptionFlag.gap_scan_req_report
    _driver_type = driver.ble_gap_opt_scan_req_report_t

    def get_opt_subfield(self, opt: driver.ble_opt_t):
        return opt.gap_opt.scan_req_report


class BleOptGapCompatMode1(BleEnableOpt):
    option_flag = BleOptionFlag.gap_compat_mode_1
    _driver_type = driver.ble_gap_opt_compat_mode_1_t

    def get_opt_subfield(self, opt: driver.ble_opt_t):
        return opt.gap_opt.compat_mode_1


class BleOptGapAuthPayloadTimeout(BleOption):
    option_flag = BleOptionFlag.gap_auth_payload_timeout

    def __init__(self, conn_handle, timeout_ms=driver.BLE_GAP_AUTH_PAYLOAD_TIMEOUT_MAX):
        self.conn_handle = conn_handle
        self.timeout_ms = timeout_ms

    def get_opt_subfield(self, opt: driver.ble_opt_t):
        return opt.gap_opt.auth_payload_timeout

    def to_c(self):
        opt = driver.ble_gap_opt_auth_payload_timeout_t()
        opt.conn_handle = self.conn_handle
        opt.auth_payload_timeout = util.msec_to_units(self.timeout_ms, util.UNIT_10_MS)
        return opt


class BleOptGapSlaveLatencyDisable(BleOption):
    option_flag = BleOptionFlag.gap_slave_latency_disable

    def __init__(self, conn_handle, disabled=False):
        self.conn_handle = conn_handle
        self.disabled = disabled

    def get_opt_subfield(self, opt: driver.ble_opt_t):
        return opt.gap_opt.slave_latency_disable

    def to_c(self):
        opt = driver.ble_gap_opt_slave_latency_disable_t()
        opt.conn_handle = self.conn_handle
        opt.disable = int(self.disabled)


class BleEnableConfig(object):
    def __init__(self,
                 vs_uuid_count=10,
                 periph_role_count=driver.BLE_GAP_ROLE_COUNT_PERIPH_DEFAULT,
                 central_role_count=driver.BLE_GAP_ROLE_COUNT_CENTRAL_DEFAULT,
                 central_sec_count=driver.BLE_GAP_ROLE_COUNT_CENTRAL_DEFAULT,
                 service_changed_char=driver.BLE_GATTS_SERVICE_CHANGED_DEFAULT,
                 attr_table_size=driver.BLE_GATTS_ATTR_TAB_SIZE_DEFAULT,
                 device_name=""):
        self.vs_uuid_count = vs_uuid_count
        self.periph_role_count = periph_role_count
        self.central_role_count = central_role_count
        self.central_sec_count = central_sec_count
        self.service_changed_char = service_changed_char
        self.attr_table_size = attr_table_size
        self.device_name = device_name

    def get_vs_uuid_cfg(self):
        config = driver.ble_cfg_t()
        cfg = config.common_cfg.vs_uuid_cfg
        cfg.vs_uuid_count = self.vs_uuid_count

        return driver.BLE_COMMON_CFG_VS_UUID, config

    def get_role_count_cfg(self):
        config = driver.ble_cfg_t()
        cfg = config.gap_cfg.role_count_cfg
        cfg.periph_role_count = self.periph_role_count
        cfg.central_role_count = self.central_role_count
        cfg.central_sec_count = self.central_sec_count

        return driver.BLE_GAP_CFG_ROLE_COUNT, config

    def get_device_name_cfg(self):
        config = driver.ble_cfg_t()
        cfg = config.gap_cfg.device_name_cfg
        cfg.current_len = len(self.device_name)
        cfg.max_len = cfg.current_len
        cfg.p_value = util.list_to_uint8_array(self.device_name).cast()
        cfg.write_perm.sm = 0
        cfg.write_perm.lv = 0
        cfg.vloc = driver.BLE_GATTS_VLOC_USER

        return driver.BLE_GAP_CFG_DEVICE_NAME, config

    def get_service_changed_cfg(self):
        config = driver.ble_cfg_t()
        cfg = config.gatts_cfg.service_changed
        cfg = driver.ble_gatts_cfg_service_changed_t()
        cfg.service_changed = self.service_changed_char

        return driver.BLE_GATTS_CFG_SERVICE_CHANGED, config

    def get_attr_tab_size_cfg(self):
        config = driver.ble_cfg_t()
        cfg = config.gatts_cfg.attr_tab_size
        cfg.attr_tab_size = self.attr_table_size

        return driver.BLE_GATTS_CFG_ATTR_TAB_SIZE, config

    def get_configs(self):
        yield self.get_vs_uuid_cfg()
        yield self.get_role_count_cfg()
        yield self.get_device_name_cfg()
        yield self.get_service_changed_cfg()
        yield self.get_attr_tab_size_cfg()


class BleConnConfig(object):
    DEFAULT_CONN_TAG = 1

    def __init__(self, conn_tag=DEFAULT_CONN_TAG,
                 conn_count=driver.BLE_GAP_CONN_COUNT_DEFAULT,
                 event_length=driver.BLE_GAP_EVENT_LENGTH_DEFAULT,
                 write_cmd_tx_queue_size=driver.BLE_GATTC_WRITE_CMD_TX_QUEUE_SIZE_DEFAULT,
                 hvn_tx_queue_size=driver.BLE_GATTS_HVN_TX_QUEUE_SIZE_DEFAULT,
                 max_att_mtu=driver.BLE_GATT_ATT_MTU_DEFAULT):  # TODO: L2CAP config
        self.conn_tag = conn_tag
        self.conn_count = conn_count
        self.event_length = event_length
        self.write_cmd_tx_queue_size = write_cmd_tx_queue_size
        self.hvn_tx_queue_size = hvn_tx_queue_size
        self.max_att_mtu = max_att_mtu

    def get_gap_config(self):
        config = driver.ble_cfg_t()
        config.conn_cfg.conn_cfg_tag = self.conn_tag
        cfg = config.conn_cfg.params.gap_conn_cfg
        cfg.conn_count = self.conn_count
        cfg.event_length = self.event_length
        return driver.BLE_CONN_CFG_GAP, config

    def get_gatt_config(self):
        config = driver.ble_cfg_t()
        config.conn_cfg.conn_cfg_tag = self.conn_tag
        cfg = config.conn_cfg.params.gatt_conn_cfg
        cfg.att_mtu = self.max_att_mtu
        return driver.BLE_CONN_CFG_GATT, config

    def get_gattc_config(self):
        config = driver.ble_cfg_t()
        config.conn_cfg.conn_cfg_tag = self.conn_tag
        cfg = config.conn_cfg.params.gattc_conn_cfg
        cfg.write_cmd_tx_queue_size = self.write_cmd_tx_queue_size
        return driver.BLE_CONN_CFG_GATTC, config

    def get_gatts_config(self):
        config = driver.ble_cfg_t()
        config.conn_cfg.conn_cfg_tag = self.conn_tag
        cfg = config.conn_cfg.params.gatts_conn_cfg
        cfg.hvn_tx_queue_size = self.hvn_tx_queue_size
        return driver.BLE_CONN_CFG_GATTS, config

    def get_configs(self):
        yield self.get_gap_config()
        yield self.get_gatt_config()
        yield self.get_gattc_config()
        yield self.get_gatts_config()



#
# Copyright (c) 2016 Nordic Semiconductor ASA
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright notice, this
#   list of conditions and the following disclaimer in the documentation and/or
#   other materials provided with the distribution.
#
#   3. Neither the name of Nordic Semiconductor ASA nor the names of other
#   contributors to this software may be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#   4. This software must only be used in or with a processor manufactured by Nordic
#   Semiconductor ASA, or in or with a processor manufactured by a third party that
#   is used in combination with a processor manufactured by Nordic Semiconductor.
#
#   5. Any software provided in binary or object form under this license must not be
#   reverse engineered, decompiled, modified and/or disassembled.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import importlib

from blatann.nrf.nrf_dll_load import driver

UNIT_0_625_MS = 625  # Unit used for scanning and advertising parameters
UNIT_1_25_MS = 1250  # Unit used for connection interval parameters
UNIT_10_MS = 10000  # Unit used for supervision timeout parameter


def msec_to_units(time_ms, resolution):
    """Convert milliseconds to BLE specific time units."""
    units = time_ms * 1000 / resolution
    return int(units)


def units_to_msec(units, resolution):
    """Convert BLE specific units to milliseconds."""
    time_ms = units * float(resolution) / 1000
    return time_ms


def char_array_to_list(array_pointer, length):
    """Convert char_array to python list."""
    data_array = driver.char_array.frompointer(array_pointer)
    data_list = _populate_list(data_array, length)
    return data_list


def uint8_array_to_list(array_pointer, length):
    """Convert uint8_array to python list."""
    data_array = driver.uint8_array.frompointer(array_pointer)
    data_list = _populate_list(data_array, length)
    return data_list


def uint16_array_to_list(array_pointer, length):
    """Convert uint16_array to python list."""
    data_array = driver.uint16_array.frompointer(array_pointer)
    data_list = _populate_list(data_array, length)
    return data_list


def service_array_to_list(array_pointer, length):
    """Convert ble_gattc_service_array to python list."""
    data_array = driver.ble_gattc_service_array.frompointer(array_pointer)
    data_list = _populate_list(data_array, length)
    return data_list


def include_array_to_list(array_pointer, length):
    """Convert ble_gattc_include_array to python list."""
    data_array = driver.ble_gattc_include_array.frompointer(array_pointer)
    data_list = _populate_list(data_array, length)
    return data_list


def ble_gattc_char_array_to_list(array_pointer, length):
    """Convert ble_gattc_char_array to python list."""
    data_array = driver.ble_gattc_char_array.frompointer(array_pointer)
    data_list = _populate_list(data_array, length)
    return data_list


def desc_array_to_list(array_pointer, length):
    """Convert ble_gattc_desc_array to python list."""
    data_array = driver.ble_gattc_desc_array.frompointer(array_pointer)
    data_list = _populate_list(data_array, length)
    return data_list


def ble_gattc_attr_info16_array_to_list(array_pointer, length):
    """Convert ble_gattc_attr_info16_array to python list"""
    data_array = driver.ble_gattc_attr_info16_array.frompointer(array_pointer)
    data_list = _populate_array(data_array, length)
    return data_list


def ble_gattc_attr_info128_array_to_list(array_pointer, length):
    """Convert ble_gattc_attr_info128_array to python list"""
    data_array = driver.ble_gattc_attr_info128_array.frompointer(array_pointer)
    data_list = _populate_array(data_array, length)
    return data_list


def handle_value_array_to_list(array_pointer, length):
    """Convert ble_gattc_handle_value_array to python list."""
    data_array = driver.ble_gattc_handle_value_array.frompointer(array_pointer)
    data_list = _populate_list(data_array, length)
    return data_list


def attr_info_array_to_list(array_pointer, length):
    """Convert ble_gattc_attr_info_array to python list."""
    data_array = driver.ble_gattc_attr_info_array.frompointer(array_pointer)
    data_list = _populate_list(data_array, length)
    return data_list


def attr_info16_array_to_list(array_pointer, length):
    """Convert ble_gattc_attr_info16_array to python list."""
    data_array = driver.ble_gattc_attr_info16_array.frompointer(array_pointer)
    data_list = _populate_list(data_array, length)
    return data_list


def attr_info128_array_to_list(array_pointer, length):
    """Convert ble_gattc_attr_info128_array to python list."""
    data_array = driver.ble_gattc_attr_info128_array.frompointer(array_pointer)
    data_list = _populate_list(data_array, length)
    return data_list


def serial_port_desc_array_to_list(array_pointer, length):
    """Convert sd_rpc_serial_port_desc_array to python list."""
    data_array = driver.sd_rpc_serial_port_desc_array.frompointer(array_pointer)
    data_list = _populate_list(data_array, length)
    return data_list


def _populate_list(data_array, length):
    data_list = []
    for i in range(0, length):
        data_list.append(data_array[i])
    return data_list


def list_to_char_array(data_list):
    """Convert python list to char_array."""

    data_array = _populate_array(data_list, driver.char_array)
    return data_array


def list_to_uint8_array(data_list):
    """Convert python list to uint8_array."""
    data_array = _populate_array(data_list, driver.uint8_array)
    return data_array


def list_to_uint16_array(data_list):
    """Convert python list to uint16_array."""
    data_array = _populate_array(data_list, driver.uint16_array)
    return data_array


def list_to_service_array(data_list):
    """Convert python list to ble_gattc_service_array."""

    data_array = _populate_array(data_list, driver.ble_gattc_service_array)
    return data_array


def list_to_include_array(data_list):
    """Convert python list to ble_gattc_include_array."""

    data_array = _populate_array(data_list, driver.ble_gattc_include_array)
    return data_array


def list_to_ble_gattc_char_array(data_list):
    """Convert python list to ble_gattc_char_array."""

    data_array = _populate_array(data_list, driver.ble_gattc_char_array)
    return data_array


def list_to_desc_array(data_list):
    """Convert python list to ble_gattc_desc_array."""

    data_array = _populate_array(data_list, driver.ble_gattc_desc_array)
    return data_array


def list_to_handle_value_array(data_list):
    """Convert python list to ble_gattc_handle_value_array."""

    data_array = _populate_array(data_list, driver.ble_gattc_handle_value_array)
    return data_array


def list_to_serial_port_desc_array(data_list):
    """Convert python list to sd_rpc_serial_port_desc_array."""

    data_array = _populate_array(data_list, driver.sd_rpc_serial_port_desc_array)
    return data_array


def _populate_array(data_list, array_type):
    length = len(data_list)
    data_array = array_type(length)
    for i in range(0, length):
        data_array[i] = data_list[i]
    return data_array



from blatann.nrf.nrf_types.enums import *
from blatann.nrf.nrf_types.generic import *
from blatann.nrf.nrf_types.gap import *
from blatann.nrf.nrf_types.smp import *
from blatann.nrf.nrf_types.gatt import *
from blatann.nrf.nrf_types.config import *



import logging

from blatann.event_type import EventSource, Event
from blatann.gatt import gattc
from blatann.nrf import nrf_events
from blatann.waitables.event_waitable import EventWaitable
from blatann.event_args import EventArgs, DatabaseDiscoveryCompleteEventArgs


logger = logging.getLogger(__name__)


class _DiscoveryEventArgs(EventArgs):
    def __init__(self, services, status):
        """
        :type services: list[gattc.GattcService]
        :type status: nrf_events.BLEGattStatusCode
        """
        self.services = services
        self.status = status


class _DiscoveryState(object):
    def __init__(self):
        self.current_handle = 0x0001
        self.services = []
        self.characteristics = []
        self.service_index = 0
        self.char_index = 0
        self.iterate_by_chars = False

    def reset(self):
        self.current_handle = 0x0001
        self.services = []
        self.characteristics = []
        self.service_index = 0
        self.char_index = 0

    @property
    def end_of_services(self):
        return self.service_index >= len(self.services)

    @property
    def end_of_characteristics(self):
        if self.iterate_by_chars:
            return self.char_index >= len(self.characteristics)
        return self.char_index >= len(self.current_service.chars)

    @property
    def current_characteristic(self):
        """
        :rtype: nrf_events.BLEGattCharacteristic
        """
        if self.iterate_by_chars:
            return self.characteristics[self.char_index]
        return self.current_service.chars[self.char_index]

    @property
    def current_service(self):
        """
        :rtype: nrf_events.BLEGattService
        """
        return self.services[self.service_index]


class _Discoverer(object):
    def __init__(self, name, ble_device, peer):
        """
        :type ble_device: blatann.BleDevice
        :type peer: blatann.peer.Peripheral
        """
        self.ble_device = ble_device
        self.peer = peer
        self._state = _DiscoveryState()
        self._on_complete_event = EventSource("{} Complete".format(name), logger)

    def start(self, services):
        raise NotImplementedError

    @property
    def on_complete(self):
        """
        :rtype: Event
        """
        return self._on_complete_event


class _ServiceDiscoverer(_Discoverer):
    def __init__(self, ble_device, peer):
        super(_ServiceDiscoverer, self).__init__("Service Discovery", ble_device, peer)

    def start(self, services=None):
        self._state.reset()

        self.ble_device.ble_driver.ble_gattc_prim_srvc_disc(self.peer.conn_handle, None,
                                                            self._state.current_handle)

        self.peer.driver_event_subscribe(self._on_primary_service_discovery, nrf_events.GattcEvtPrimaryServiceDiscoveryResponse)
        self.peer.driver_event_subscribe(self._on_service_uuid_read, nrf_events.GattcEvtReadResponse)
        return EventWaitable(self.on_complete)

    def _on_complete(self, status=nrf_events.BLEGattStatusCode.success):
        self.peer.driver_event_unsubscribe(self._on_primary_service_discovery)
        self.peer.driver_event_unsubscribe(self._on_service_uuid_read)
        self._on_complete_event.notify(self, _DiscoveryEventArgs(self._state.services, status))

    def _on_primary_service_discovery(self, driver, event):
        """
        :type event: nrf_events.GattcEvtPrimaryServiceDiscoveryResponse
        """
        if not self.peer.connected:
            logger.warning("Primary service discovery for a disconnected peer")
        if event.conn_handle != self.peer.conn_handle:
            return

        if event.status != nrf_events.BLEGattStatusCode.success:
            self._on_complete(event.status)  # Not found, done
            return

        self._state.services.extend(event.services)
        end_handle = event.services[-1].end_handle
        if end_handle != 0xFFFF:
            # Continue service discovery
            self.ble_device.ble_driver.ble_gattc_prim_srvc_disc(self.peer.conn_handle, None,
                                                                end_handle+1)
            return

        # Done discovering services, now discover their attributes
        self._discover_uuids()

    def _discover_uuids(self):
        while not self._state.end_of_services:
            service = self._state.current_service
            if service.uuid.base.type == 0:  # Unknown base, register
                self.ble_device.ble_driver.ble_gattc_read(self.peer.conn_handle, service.start_handle)
                return
            self._state.service_index += 1
        self._on_complete()

    def _on_service_uuid_read(self, driver, event):
        """
        :type event: nrf_events.GattcEvtReadResponse
        """
        if self.peer.conn_handle != event.conn_handle:
            return

        logger.info(("Got gattc read: {}".format(event)))
        service = self._state.current_service

        if event.attr_handle != service.start_handle:
            return

        # Length should be 16 for 128-bit uuids
        if len(event.data) != 16:
            logger.error("Service UUID not 16 bytes: {}".format(event.data))
        else:
            nrf_uuid = nrf_events.BLEUUID.from_array(event.data)
            self.ble_device.uuid_manager.register_uuid(nrf_uuid)
            logger.info("Discovered UUID: {}".format(nrf_uuid))
            self._state.current_service.uuid = nrf_uuid

        self._state.service_index += 1
        self._discover_uuids()


class _CharacteristicDiscoverer(_Discoverer):
    def __init__(self, ble_device, peer):
        """
        :type ble_device: blatann.device.BleDevice
        :type peer: blatann.peer.Peripheral
        """
        super(_CharacteristicDiscoverer, self).__init__("Characteristic Discovery", ble_device, peer)

    def start(self, services):
        self._state.reset()
        self._state.services = services

        self.peer.driver_event_subscribe(self._on_characteristic_discovery, nrf_events.GattcEvtCharacteristicDiscoveryResponse)
        self.peer.driver_event_subscribe(self._on_char_uuid_read, nrf_events.GattcEvtReadResponse)

        self._discover_characteristics()
        return EventWaitable(self.on_complete)

    def _on_complete(self, status=nrf_events.BLEGattStatusCode.success):
        self.peer.driver_event_unsubscribe(self._on_characteristic_discovery)
        self.peer.driver_event_unsubscribe(self._on_char_uuid_read)
        self._on_complete_event.notify(self, _DiscoveryEventArgs(self._state.services, status))

    def _discover_characteristics(self):
        service = self._state.current_service
        self.ble_device.ble_driver.ble_gattc_char_disc(self.peer.conn_handle, service.start_handle, service.end_handle)

    def _on_characteristic_discovery(self, driver, event):
        """
        :type event: nrf_events.GattcEvtCharacteristicDiscoveryResponse
        """
        if not self.peer.connected:
            logger.warning("Primary service discovery for a disconnected peer")
        if event.conn_handle != self.peer.conn_handle:
            return

        if event.status == nrf_events.BLEGattStatusCode.attribute_not_found:
            # Done discovering characteristics in this service, discover next service or unknown UUIDs
            self._state.service_index += 1
            self._state.char_index = 0
            if self._state.end_of_services:
                self._state.service_index = 0
                self._discover_uuids()
            else:
                self._discover_characteristics()
            return
        elif event.status != nrf_events.BLEGattStatusCode.success:
            self._on_complete(event.status)
            return

        service = self._state.current_service
        for c in event.characteristics:
            service.char_add(c)
        last_char = event.characteristics[-1]
        if last_char.handle_value == service.end_handle:
            self._state.service_index += 1
            self._state.char_index = 0
            if self._state.end_of_services:
                self._state.service_index = 0
                self._discover_uuids()
            else:
                self._discover_characteristics()
            return
        else:
            self.ble_device.ble_driver.ble_gattc_char_disc(self.peer.conn_handle, last_char.handle_decl + 1, service.end_handle)

    def _discover_uuids(self):
        while not self._state.end_of_services:
            while not self._state.end_of_characteristics:
                # Check uuid
                char = self._state.current_characteristic
                if char.uuid.base.type == 0:
                    self.ble_device.ble_driver.ble_gattc_read(self.peer.conn_handle, char.handle_decl)
                    return
                self._state.char_index += 1
            self._state.char_index = 0
            self._state.service_index += 1
        self._on_complete()

    def _on_char_uuid_read(self, driver, event):
        """
        :type event: nrf_events.GattcEvtReadResponse
        """
        if self.peer.conn_handle != event.conn_handle:
            return

        logger.info(("Got gattc read: {}".format(event)))
        char = self._state.current_characteristic

        if event.attr_handle != char.handle_decl:
            return

        # First 3 bytes are [permissions (1 byte), value handle (2 bytes)]
        uuid_bytes = event.data[3:]
        # Length should be 16 for 128-bit uuids
        if len(uuid_bytes) != 16:
            logger.error("Characteristic UUID not 16 bytes: {}".format(uuid_bytes))
        else:
            nrf_uuid = nrf_events.BLEUUID.from_array(uuid_bytes)
            self.ble_device.uuid_manager.register_uuid(nrf_uuid)
            logger.info("Discovered UUID: {}".format(nrf_uuid))
            char.uuid = nrf_uuid

        self._state.char_index += 1
        if self._state.end_of_characteristics:
            self._state.service_index += 1
            self._state.char_index = 0
        self._discover_uuids()


class _DescriptorDiscoverer(_Discoverer):
    def __init__(self, ble_device, peer):
        """
        :type ble_device: blatann.device.BleDevice
        :type peer: blatann.peer.Peripheral
        """
        super(_DescriptorDiscoverer, self).__init__("Descriptor Discovery", ble_device, peer)
        self._state.iterate_by_chars = True

    def start(self, services):
        self._state.reset()
        self._state.services = services
        # Compile the characteristics into a single list so its easier to iterate
        for s in services:
            self._state.characteristics.extend(s.chars)
        self.peer.driver_event_subscribe(self._on_descriptor_discovery, nrf_events.GattcEvtDescriptorDiscoveryResponse)
        on_complete_waitable = EventWaitable(self.on_complete)
        if not self._state.characteristics:
            self._on_complete()
        else:
            self._discover_descriptors()
        return on_complete_waitable

    def _on_complete(self, status=nrf_events.BLEGattStatusCode.success):
        self.peer.driver_event_unsubscribe(self._on_descriptor_discovery)
        self._on_complete_event.notify(self, _DiscoveryEventArgs(self._state.services, status))

    def _discover_descriptors(self, starting_handle=None):
        char = self._state.current_characteristic
        if starting_handle is None:
            starting_handle = char.handle_value
        self.peer.driver_event_subscribe(self._on_descriptor_discovery, nrf_events.GattcEvtDescriptorDiscoveryResponse)
        self.ble_device.ble_driver.ble_gattc_desc_disc(self.peer.conn_handle, starting_handle, char.end_handle)

    def _on_descriptor_discovery(self, driver, event):
        """
        :type event: nrf_events.GattcEvtDescriptorDiscoveryResponse
        """
        if not self.peer.connected:
            logger.warning("Primary service discovery for a disconnected peer")
        if event.conn_handle != self.peer.conn_handle:
            return

        if event.status == nrf_events.BLEGattStatusCode.attribute_not_found:
            self._state.char_index += 1
            if self._state.end_of_characteristics:
                self._on_complete()
                return
        elif event.status != nrf_events.BLEGattStatusCode.success:
            self._on_complete(event.status)
            return

        char = self._state.current_characteristic
        char.descs.extend(event.descriptions)

        last_desc = event.descriptions[-1]
        if last_desc.handle == char.end_handle:
            self._state.char_index += 1
            if self._state.end_of_characteristics:
                self._on_complete()
                return
        self._discover_descriptors(last_desc.handle+1)


class DatabaseDiscoverer(object):
    def __init__(self, ble_device, peer):
        """
        :type ble_device: blatann.device.BleDevice
        :type peer: blatann.peer.Peripheral
        """
        self.ble_device = ble_device
        self.peer = peer
        self._on_discovery_complete = EventSource("Service Discovery Complete", logger)
        self._on_database_discovery_complete = EventSource("Service Discovery Complete", logger)
        self._state = _DiscoveryState()
        self._service_discoverer = _ServiceDiscoverer(ble_device, peer)
        self._characteristic_discoverer = _CharacteristicDiscoverer(ble_device, peer)
        self._descriptor_discoverer = _DescriptorDiscoverer(ble_device, peer)

    @property
    def on_discovery_complete(self):
        """
        :rtype: Event
        """
        return self._on_discovery_complete

    def _on_service_discovery_complete(self, sender, event_args):
        """
        :type sender: _ServiceDiscoverer
        :type event_args: _DiscoveryEventArgs
        """
        logger.info("Service Discovery complete")
        if event_args.status != nrf_events.BLEGattStatusCode.success:
            logger.error("Error discovering services: {}".format(event_args.status))
            self._on_complete([], event_args.status)
        else:
            self._characteristic_discoverer.start(event_args.services).then(self._on_characteristic_discovery_complete)

    def _on_characteristic_discovery_complete(self, sender, event_args):
        """
        :type sender: _CharacteristicDiscoverer
        :type event_args: _DiscoveryEventArgs
        """
        logger.info("Characteristic Discovery complete")
        if event_args.status != nrf_events.BLEGattStatusCode.success:
            logger.error("Error discovering characteristics: {}".format(event_args.status))
            self._on_complete([], event_args.status)
        else:
            self._descriptor_discoverer.start(event_args.services).then(self._on_descriptor_discovery_complete)

    def _on_descriptor_discovery_complete(self, sender, event_args):
        """
        :type sender: _DescriptorDiscoverer
        :type event_args: _DiscoveryEventArgs
        """
        logger.info("Descriptor Discovery complete")
        self._on_complete(event_args.services, event_args.status)

    def _on_complete(self, services, status):
        self.peer.database.add_discovered_services(services)
        self._on_discovery_complete.notify(self.peer, DatabaseDiscoveryCompleteEventArgs(status))
        logger.info("Database Discovery complete!!")

    def start(self):
        logger.info("Starting discovery..")
        self._service_discoverer.start().then(self._on_service_discovery_complete)



from blatann.uuid import Uuid128

NORDIC_UART_SERVICE_UUID = Uuid128("6E400001-B5A3-F393-E0A9-E50E24DCCA9E")

# TX and RX characteristics are from the perspective of the client. The server will receive on TX and transmit on RX
NORDIC_UART_TX_CHARACTERISTIC_UUID = NORDIC_UART_SERVICE_UUID.new_uuid_from_base(0x0002)
NORDIC_UART_RX_CHARACTERISTIC_UUID = NORDIC_UART_SERVICE_UUID.new_uuid_from_base(0x0003)
# Slight deviation from Nordic's implementation so the client can read the server's characteristic size
NORDIC_UART_FEATURE_CHARACTERISTIC_UUID = NORDIC_UART_SERVICE_UUID.new_uuid_from_base(0x0004)



"""
This example demonstrates using Bluetooth SIG's defined Glucose service as a peripheral.
The peripheral creates a range of fake glucose readings that can be queried from the central.

This can be used in conjunction with the nRF Connect apps to explore the peripheral's functionality
"""
import logging
import datetime

from blatann import BleDevice
from blatann.gap import advertising, IoCapabilities
from blatann.utils import setup_logger
from blatann.services import glucose
from blatann.services.glucose import GlucoseFeatureType
from blatann.waitables import GenericWaitable


logger = setup_logger(level="INFO")


def on_connect(peer, event_args):
    """
    Event callback for when a central device connects to us

    :param peer: The peer that connected to us
    :type peer: blatann.peer.Client
    :param event_args: None
    """
    if peer:
        logger.info("Connected to {} peer".format("previously-bonded" if peer.is_previously_bonded else "new"))
    else:
        logger.warning("Connection timed out")


def on_disconnect(peer, event_args):
    """
    Event callback for when the client disconnects from us (or when we disconnect from the client)

    :param peer: The peer that disconnected
    :type peer: blatann.peer.Client
    :param event_args: The event args
    :type event_args: blatann.event_args.DisconnectionEventArgs
    """
    logger.info("Disconnected from peer, reason: {}".format(event_args.reason))


def on_security_level_changed(peer, event_args):
    """
    Event callback for when the security level changes on a connection with a peer

    :param peer: The peer the security level changed on
    :type peer: blatann.peer.Client
    :param event_args: The event args
    :type event_args: blatann.event_args.SecurityLevelChangedEventArgs
    """
    logger.info("Security level changed to {}".format(event_args.security_level))


def display_passkey(peer, event_args):
    """
    Event callback that is called when a passkey is required to be displayed to a user
    for the pairing process.

    :param peer: The peer the passkey is for
    :type peer: blatann.peer.Client
    :param event_args: The event args
    :type event_args: blatann.event_args.PasskeyDisplayEventArgs
    """
    if not event_args.match_request:
        logger.info("Passkey: {}".format(event_args.passkey))
    else:
        response = input("Passkey: {}, do both devices show same passkey? [y/n]\n".format(event_args.passkey))
        match_confirmed = response.lower().startswith("y")
        event_args.match_confirm(match_confirmed)


def add_fake_glucose_readings(glucose_database, num_records=15):
    """
    Helper method to create some glucose readings and add them to the glucose database

    :param glucose_database: The database to add readings to
    :type glucose_database: glucose.BasicGlucoseDatabase
    :param num_records: The number of records to generate
    """
    init_time = datetime.datetime.now()

    for i in range(0, num_records):
        # Increment the reading times by 5 mins
        sample_time = init_time + datetime.timedelta(minutes=i * 5)

        # create a sample reading with some random data
        sample = glucose.GlucoseSample(glucose.GlucoseType.capillary_plasma, glucose.SampleLocation.finger,
                                       12.345 * i, glucose.GlucoseConcentrationUnits.mol_per_liter)

        # Create the measurement
        m = glucose.GlucoseMeasurement(i, sample_time, sample=sample)

        # Add some records with context
        if i % 4 == 0:
            carbs = glucose.CarbsInfo(carbs_grams=50 * i, carb_type=glucose.CarbohydrateType.lunch)
            medication = glucose.MedicationInfo(glucose.MedicationType.long_acting_insulin, 5.41 * i,
                                                glucose.MedicationUnits.milligrams)
            context = glucose.GlucoseContext(i, carbs=carbs, medication=medication, hba1c_percent=i * 6.05)
            # Add the context to the measurement
            m.context = context

        glucose_database.add_record(m)


def main(serial_port):
    # Create and open the device
    ble_device = BleDevice(serial_port)
    ble_device.open()

    # Create a database to store the readings
    glucose_database = glucose.BasicGlucoseDatabase()
    # Add the service to the BLE database, using the glucose database just created, require encryption at the minimum
    service = glucose.add_glucose_service(ble_device.database, glucose_database, glucose.SecurityLevel.MITM)

    # Set the features of this "glucose sensor"
    features = glucose.GlucoseFeatures(GlucoseFeatureType.low_battery_detection, GlucoseFeatureType.strip_insertion_error_detection)
    service.set_features(features)

    # Add some measurements to the glucose database
    add_fake_glucose_readings(glucose_database)

    # Register listeners for when the client connects and disconnects
    ble_device.client.on_connect.register(on_connect)
    ble_device.client.on_disconnect.register(on_disconnect)

    # Set the connection parameters for the client
    ble_device.client.set_connection_parameters(15, 100, 4000)

    # Set the function to display the passkey
    ble_device.client.security.on_passkey_display_required.register(display_passkey)

    # Add a callback for when the security level changes
    ble_device.client.security.on_security_level_changed.register(on_security_level_changed)

    # Set the security parameters for the client
    ble_device.client.security.set_security_params(passcode_pairing=False, bond=True, lesc_pairing=True,
                                                   io_capabilities=IoCapabilities.KEYBOARD_DISPLAY, out_of_band=False)

    # Advertise the Glucose service
    adv_data = advertising.AdvertisingData(local_name="Glucose Test", flags=0x06,
                                           service_uuid16s=glucose.GLUCOSE_SERVICE_UUID)
    ble_device.advertiser.set_advertise_data(adv_data)

    logger.info("Advertising")
    ble_device.advertiser.start(timeout_sec=0, auto_restart=True)

    # Create a waitable that will never fire, and wait for some time
    w = GenericWaitable()
    w.wait(60*30, exception_on_timeout=False)  # Keep device active for 30 mins

    logger.info("Done")
    ble_device.close()


if __name__ == '__main__':
    main("COM8")



from __future__ import annotations
import logging
import threading
from typing import Callable, List, Optional, Iterable

from blatann import gatt
from blatann.uuid import Uuid
from blatann.event_type import EventSource, Event
from blatann.gatt.reader import GattcReader
from blatann.gatt.writer import GattcWriter
from blatann.nrf import nrf_types, nrf_events
from blatann.waitables.event_waitable import EventWaitable, IdBasedEventWaitable
from blatann.exceptions import InvalidOperationException, InvalidStateException
from blatann.event_args import *
from blatann.utils.queued_tasks_manager import QueuedTasksManagerBase

logger = logging.getLogger(__name__)


class GattcCharacteristic(gatt.Characteristic):
    def __init__(self, ble_device, peer, read_write_manager, uuid, properties,
                 declaration_handle, value_handle, cccd_handle=None):
        """
        :type ble_device: blatann.BleDevice
        :type peer: blatann.peer.Peripheral
        :type read_write_manager: _ReadWriteManager
        :type uuid: blatann.uuid.Uuid
        :type properties: gatt.CharacteristicProperties
        :param declaration_handle:
        :param value_handle:
        :param cccd_handle:
        """
        super(GattcCharacteristic, self).__init__(ble_device, peer, uuid, properties)
        self.declaration_handle = declaration_handle
        self.value_handle = value_handle
        self.cccd_handle = cccd_handle
        self._manager = read_write_manager
        self._value = b""

        self._on_notification_event = EventSource("On Notification", logger)
        self._on_read_complete_event = EventSource("On Read Complete", logger)
        self._on_write_complete_event = EventSource("Write Complete", logger)
        self._on_cccd_write_complete_event = EventSource("CCCD Write Complete", logger)

        self.peer.driver_event_subscribe(self._on_indication_notification, nrf_events.GattcEvtHvx)
        self._manager.on_write_complete.register(self._write_complete)
        self._manager.on_read_complete.register(self._read_complete)

    """
    Properties
    """

    @property
    def value(self) -> bytes:
        """
        The current value of the characteristic

        :return: The last known value of the characteristic
        """
        return self._value

    @property
    def readable(self) -> bool:
        """
        Gets if the characteristic can be read from
        """
        return self._properties.read

    @property
    def writable(self) -> bool:
        """
        Gets if the characteristic can be written to
        """
        return self._properties.write

    @property
    def subscribable(self) -> bool:
        """
        Gets if the characteristic can be subscribed to
        """
        return self._properties.notify or self._properties.indicate

    @property
    def subscribed(self) -> bool:
        """
        Gets if the characteristic is currently subscribed to
        """
        return self.cccd_state != gatt.SubscriptionState.NOT_SUBSCRIBED

    """
    Events
    """

    @property
    def on_read_complete(self) -> Event[GattcCharacteristic, ReadCompleteEventArgs]:
        return self._on_read_complete_event

    @property
    def on_write_complete(self) -> Event[GattcCharacteristic, WriteCompleteEventArgs]:
        return self._on_write_complete_event

    """
    Public Methods
    """

    def subscribe(self, on_notification_handler: Callable[[GattcCharacteristic, NotificationReceivedEventArgs], None],
                  prefer_indications=False) -> EventWaitable[GattcCharacteristic, SubscriptionWriteCompleteEventArgs]:
        """
        Subscribes to the characteristic's indications or notifications, depending on what's available and the
        prefer_indications setting. Returns a Waitable that executes when the subscription on the peripheral finishes.

        The Waitable returns two parameters: (GattcCharacteristic this, SubscriptionWriteCompleteEventArgs event args)

        :param on_notification_handler: The handler to be called when an indication or notification is received from
            the peripheral. Must take two parameters: (GattcCharacteristic this, NotificationReceivedEventArgs event args)
        :param prefer_indications: If the peripheral supports both indications and notifications,
            will subscribe to indications instead of notifications
        :return: A Waitable that will fire when the subscription finishes
        :raises: InvalidOperationException if the characteristic cannot be subscribed to
            (characteristic does not support indications or notifications)
        """
        if not self.subscribable:
            raise InvalidOperationException("Cannot subscribe to Characteristic {}".format(self.uuid))
        if prefer_indications and self._properties.indicate or not self._properties.notify:
            value = gatt.SubscriptionState.INDICATION
        else:
            value = gatt.SubscriptionState.NOTIFY
        self._on_notification_event.register(on_notification_handler)

        write_id = self._manager.write(self.cccd_handle, gatt.SubscriptionState.to_buffer(value))
        return IdBasedEventWaitable(self._on_cccd_write_complete_event, write_id)

    def unsubscribe(self) -> EventWaitable[GattcCharacteristic, SubscriptionWriteCompleteEventArgs]:
        """
        Unsubscribes from indications and notifications from the characteristic and clears out all handlers
        for the characteristic's on_notification event handler. Returns a Waitable that executes when the unsubscription
        finishes.

        The Waitable returns two parameters: (GattcCharacteristic this, SubscriptionWriteCompleteEventArgs event args)

        :return: A Waitable that will fire when the unsubscription finishes
        """
        if not self.subscribable:
            raise InvalidOperationException("Cannot subscribe to Characteristic {}".format(self.uuid))
        value = gatt.SubscriptionState.NOT_SUBSCRIBED
        write_id = self._manager.write(self.cccd_handle, gatt.SubscriptionState.to_buffer(value))
        self._on_notification_event.clear_handlers()

        return IdBasedEventWaitable(self._on_cccd_write_complete_event, write_id)

    def read(self) -> EventWaitable[GattcCharacteristic, ReadCompleteEventArgs]:
        """
        Initiates a read of the characteristic and returns a Waitable that executes when the read finishes with
        the data read.

        The Waitable returns two parameters: (GattcCharacteristic this, ReadCompleteEventArgs event args)

        :return: A waitable that will fire when the read finishes
        :raises: InvalidOperationException if characteristic not readable
        """
        if not self.readable:
            raise InvalidOperationException("Characteristic {} is not readable".format(self.uuid))
        read_id = self._manager.read(self.value_handle)
        return IdBasedEventWaitable(self._on_read_complete_event, read_id)

    def write(self, data) -> EventWaitable[GattcCharacteristic, WriteCompleteEventArgs]:
        """
        Initiates a write of the data provided to the characteristic and returns a Waitable that executes
        when the write completes.

        The Waitable returns two parameters: (GattcCharacteristic this, WriteCompleteEventArgs event args)

        :param data: The data to write. Can be a string, bytes, or anything that can be converted to bytes
        :type data: str or bytes or bytearray
        :return: A waitable that returns when the write finishes
        :raises: InvalidOperationException if characteristic is not writable
        """
        if isinstance(data, str):
            data = data.encode(self.string_encoding)

        if not self.writable:
            raise InvalidOperationException("Characteristic {} is not writable".format(self.uuid))
        write_id = self._manager.write(self.value_handle, bytes(data))
        return IdBasedEventWaitable(self._on_write_complete_event, write_id)

    """
    Event Handlers
    """

    def _read_complete(self, sender, event_args):
        """
        Handler for _ReadWriteManager.on_read_complete.
        Dispatches the on_read_complete event and updates the internal value if read was successful

        :param sender: The reader that the read completed on
        :type sender: _ReadWriteManager
        :param event_args: The event arguments
        :type event_args: _ReadTask
        """
        if event_args.handle == self.value_handle:
            if event_args.status == nrf_types.BLEGattStatusCode.success:
                self._value = event_args.data
            args = ReadCompleteEventArgs(event_args.id, self._value, event_args.status, event_args.reason)
            self._on_read_complete_event.notify(self, args)

    def _write_complete(self, sender, event_args):
        """
        Handler for _ReadWriteManager.on_write_complete. Dispatches on_write_complete or on_cccd_write_complete
        depending on the handle the write finished on.

        :param sender: The writer that the write completed on
        :type sender: _ReadWriteManager
        :param event_args: The event arguments
        :type event_args: _WriteTask
        """
        # Success, update the local value
        if event_args.handle == self.value_handle:
            if event_args.status == nrf_types.BLEGattStatusCode.success:
                self._value = event_args.data
            args = WriteCompleteEventArgs(event_args.id, self._value, event_args.status, event_args.reason)
            self._on_write_complete_event.notify(self, args)
        elif event_args.handle == self.cccd_handle:
            if event_args.status == nrf_types.BLEGattStatusCode.success:
                self.cccd_state = gatt.SubscriptionState.from_buffer(bytearray(event_args.data))
            args = SubscriptionWriteCompleteEventArgs(event_args.id, self.cccd_state,
                                                      event_args.status, event_args.reason)
            self._on_cccd_write_complete_event.notify(self, args)

    def _on_indication_notification(self, driver, event):
        """
        Handler for GattcEvtHvx. Dispatches the on_notification_event to listeners

        :type event: nrf_events.GattcEvtHvx
        """
        if event.conn_handle != self.peer.conn_handle or event.attr_handle != self.value_handle:
            return

        is_indication = False
        if event.hvx_type == nrf_events.BLEGattHVXType.indication:
            is_indication = True
            self.ble_device.ble_driver.ble_gattc_hv_confirm(event.conn_handle, event.attr_handle)
        self._value = bytearray(event.data)
        self._on_notification_event.notify(self, NotificationReceivedEventArgs(self._value, is_indication))

    """
    Factory methods
    """

    @classmethod
    def from_discovered_characteristic(cls, ble_device, peer, read_write_manager, nrf_characteristic):
        """
        Internal factory method used to create a new characteristic from a discovered nRF Characteristic

        :type nrf_characteristic: nrf_types.BLEGattCharacteristic
        """
        char_uuid = ble_device.uuid_manager.nrf_uuid_to_uuid(nrf_characteristic.uuid)
        properties = gatt.CharacteristicProperties.from_nrf_properties(nrf_characteristic.char_props)
        cccd_handle_list = [d.handle for d in nrf_characteristic.descs
                            if d.uuid == nrf_types.BLEUUID.Standard.cccd]
        cccd_handle = cccd_handle_list[0] if cccd_handle_list else None
        return GattcCharacteristic(ble_device, peer, read_write_manager, char_uuid, properties,
                                   nrf_characteristic.handle_decl, nrf_characteristic.handle_value, cccd_handle)


class GattcService(gatt.Service):
    @property
    def characteristics(self) -> List[GattcCharacteristic]:
        """
        Gets the list of characteristics within the service

        :rtype: list of GattcCharacteristic
        """
        return self._characteristics

    def find_characteristic(self, characteristic_uuid: Uuid) -> Optional[GattcCharacteristic]:
        """
        Finds the characteristic matching the given UUID inside the service. If not found, returns None

        :param characteristic_uuid: The UUID of the characteristic to find
        :return: The characteristic if found, otherwise None
        """
        for c in self.characteristics:
            if c.uuid == characteristic_uuid:
                return c

    @classmethod
    def from_discovered_service(cls, ble_device, peer, read_write_manager, nrf_service):
        """
        Internal factory method used to create a new service from a discovered nRF Service.
        Also takes care of creating and adding all characteristics within the service

        :type ble_device: blatann.device.BleDevice
        :type peer: blatann.peer.Peripheral
        :type read_write_manager: _ReadWriteManager
        :type nrf_service: nrf_types.BLEGattService
        """
        service_uuid = ble_device.uuid_manager.nrf_uuid_to_uuid(nrf_service.uuid)
        service = GattcService(ble_device, peer, service_uuid, gatt.ServiceType.PRIMARY,
                               nrf_service.start_handle, nrf_service.end_handle)
        for c in nrf_service.chars:
            char = GattcCharacteristic.from_discovered_characteristic(ble_device, peer, read_write_manager, c)
            service.characteristics.append(char)
        return service


class GattcDatabase(gatt.GattDatabase):
    """
    Represents a remote GATT Database which lives on a connected peripheral. Contains all discovered services,
    characteristics, and descriptors
    """
    def __init__(self, ble_device, peer):
        super(GattcDatabase, self).__init__(ble_device, peer)
        self._writer = GattcWriter(ble_device, peer)
        self._reader = GattcReader(ble_device, peer)
        self._read_write_manager = _ReadWriteManager(self._reader, self._writer)

    @property
    def services(self) -> List[GattcService]:
        """
        :rtype: list of GattcService
        """
        return self._services

    def find_service(self, service_uuid: Uuid) -> Optional[GattcService]:
        """
        Finds the characteristic matching the given UUID inside the database. If not found, returns None

        :param service_uuid: The UUID of the service to find
        :type service_uuid: blatann.uuid.Uuid
        :return: The service if found, otherwise None
        :rtype: GattcService
        """
        for s in self.services:
            if s.uuid == service_uuid:
                return s

    def find_characteristic(self, characteristic_uuid):
        """
        Finds the characteristic matching the given UUID inside the database. If not found, returns None

        :param characteristic_uuid: The UUID of the characteristic to find
        :type characteristic_uuid: blatann.uuid.Uuid
        :return: The characteristic if found, otherwise None
        :rtype: GattcCharacteristic
        """
        for c in self.iter_characteristics():
            if c.uuid == characteristic_uuid:
                return c

    def iter_characteristics(self) -> Iterable[GattcCharacteristic]:
        """
        Iterates through all the characteristics in the database

        :return: An iterable of the characterisitcs in the database
        :rtype: collections.Iterable[GattcCharacteristic]
        """
        for s in self.services:
            for c in s.characteristics:
                yield c

    def add_discovered_services(self, nrf_services):
        """
        Adds the discovered NRF services from the service_discovery module.
        Used for internal purposes primarily.

        :param nrf_services: The discovered services with all the characteristics and descriptors
        :type nrf_services: list of nrf_types.BLEGattService
        """
        for service in nrf_services:
            self.services.append(GattcService.from_discovered_service(self.ble_device, self.peer,
                                                                      self._read_write_manager, service))


class _ReadTask(object):
    _id_counter = 1
    _lock = threading.Lock()

    def __init__(self, handle):
        with _ReadTask._lock:
            self.id = _ReadTask._id_counter
            _ReadTask._id_counter += 1
        self.handle = handle
        self.data = ""
        self.status = gatt.GattStatusCode.unknown
        self.reason = GattOperationCompleteReason.FAILED


class _WriteTask(object):
    _id_counter = 1
    _lock = threading.Lock()

    def __init__(self, handle, data):
        with _WriteTask._lock:
            self.id = _WriteTask._id_counter
            _WriteTask._id_counter += 1
        self.handle = handle
        self.data = data
        self.status = gatt.GattStatusCode.unknown
        self.reason = GattOperationCompleteReason.FAILED


class _ReadWriteManager(QueuedTasksManagerBase):
    def __init__(self, reader, writer):
        """
        :type reader: GattcReader
        :type writer: GattcWriter
        """
        super(_ReadWriteManager, self).__init__()
        self._reader = reader
        self._writer = writer
        self._reader.peer.on_disconnect.register(self._on_disconnect)
        self._cur_read_task = None
        self._cur_write_task = None
        self.on_read_complete = EventSource("Gattc Read Complete", logger)
        self.on_write_complete = EventSource("Gattc Write Complete", logger)
        self._reader.on_read_complete.register(self._read_complete)
        self._writer.on_write_complete.register(self._write_complete)
        self._reader.peer.driver_event_subscribe(self._on_timeout, nrf_events.GattcEvtTimeout)

    def read(self, handle):
        read_task = _ReadTask(handle)
        self._add_task(read_task)
        return read_task.id

    def write(self, handle, value):
        write_task = _WriteTask(handle, value)
        self._add_task(write_task)
        return write_task.id

    def clear_all(self):
        self._clear_all(GattOperationCompleteReason.QUEUE_CLEARED)

    def _handle_task(self, task):
        if isinstance(task, _ReadTask):
            self._reader.read(task.handle)
            self._cur_read_task = task
        elif isinstance(task, _WriteTask):
            self._writer.write(task.handle, task.data)
            self._cur_write_task = task
        else:
            return True

    def _handle_task_failure(self, task, e):
        if isinstance(task, _ReadTask):
            self.on_read_complete.notify(self, task)
        elif isinstance(task, _WriteTask):
            self.on_write_complete.notify(self, task)

    def _handle_task_cleared(self, task, reason):
        if isinstance(task, _ReadTask):
            task.reason = reason
            self.on_read_complete.notify(self, task)
        elif isinstance(task, _WriteTask):
            task.reason = reason
            self.on_write_complete.notify(self, task)

    def _on_disconnect(self, sender, event_args):
        self._clear_all(GattOperationCompleteReason.SERVER_DISCONNECTED)

    def _on_timeout(self, sender, event_args):
        self._clear_all(GattOperationCompleteReason.TIMEOUT)

    def _read_complete(self, sender, event_args):
        """
        Handler for GattcReader.on_read_complete.
        Dispatches the on_read_complete event and updates the internal value if read was successful

        :param sender: The reader that the read completed on
        :type sender: blatann.gatt.reader.GattcReader
        :param event_args: The event arguments
        :type event_args: blatann.gatt.reader.GattcReadCompleteEventArgs
        """
        task = self._cur_read_task
        self._task_completed(self._cur_read_task)

        task.data = event_args.data
        task.status = event_args.status
        self.on_read_complete.notify(sender, task)

    def _write_complete(self, sender, event_args):
        """
        Handler for GattcWriter.on_write_complete. Dispatches on_write_complete or on_cccd_write_complete
        depending on the handle the write finished on.

        :param sender: The writer that the write completed on
        :type sender: blatann.gatt.writer.GattcWriter
        :param event_args: The event arguments
        :type event_args: blatann.gatt.writer.GattcWriteCompleteEventArgs
        """
        task = self._cur_write_task
        self._task_completed(self._cur_write_task)

        task.status = event_args.status
        self.on_write_complete.notify(sender, task)



from blatann.uuid import Uuid16

CURRENT_TIME_SERVICE_UUID = Uuid16(0x1805)

CURRENT_TIME_CHARACTERISTIC_UUID = Uuid16(0x2A2B)
LOCAL_TIME_INFO_CHARACTERISTIC_UUID = Uuid16(0x2A0F)
REFERENCE_INFO_CHARACTERISTIC_UUID = Uuid16(0x2A14)



from enum import IntEnum
import math
import struct
import datetime


class BleDataStream(object):
    def __init__(self, value=b""):
        self.value = value
        self.decode_index = 0

    def __repr__(self):
        return repr(self.value)

    def __str__(self):
        return repr(self.value)

    def __getitem__(self, item):
        return self.value[item]

    def __len__(self):
        return len(self.value) - self.decode_index

    def encode(self, ble_type, *values):
        stream = ble_type.encode(*values)
        if isinstance(stream, BleDataStream):
            self.value += stream.value
        else:
            self.value += stream

    def encode_multiple(self, *ble_type_value_pairs):
        for type_values in ble_type_value_pairs:
            self.encode(type_values[0], *type_values[1:])

    def encode_if(self, conditional, ble_type, *values):
        if conditional:
            self.encode(ble_type, *values)

    def encode_if_multiple(self, conditional, *ble_type_value_pairs):
        if conditional:
            self.encode_multiple(*ble_type_value_pairs)

    def decode(self, ble_type):
        return ble_type.decode(self)

    def decode_if(self, conditional, ble_type):
        if conditional:
            return self.decode(ble_type)

    def decode_multiple(self, *ble_types):
        return map(self.decode, ble_types)

    def decode_if_multiple(self, conditional, *ble_types):
        if conditional:
            return self.decode_multiple(*ble_types)
        return [None] * len(ble_types)

    def take(self, num_bytes):
        s = self.value[self.decode_index:self.decode_index + num_bytes]
        self.decode_index += num_bytes
        return s

    def take_all(self):
        s = self.value[self.decode_index:]
        self.decode_index = len(self.value)
        return s


class BleCompoundDataType(object):
    data_stream_types = []

    def encode_values(self, *values):
        """
        :rtype: BleDataStream
        """
        stream = BleDataStream()
        for value, data_type in zip(values, self.data_stream_types):
            stream.encode(data_type, value)
        return stream

    def encode(self):
        """
        :rtype: BleDataStream
        """
        raise NotImplementedError()

    @classmethod
    def decode(cls, stream):
        """
        :type stream: BleDataStream
        :return: The values decoded from the stream
        :rtype: tuple
        """
        values = []
        for data_type in cls.data_stream_types:
            value = stream.decode(data_type)
            values.append(value)
        return values

    @classmethod
    def encoded_size(cls):
        size = 0
        for t in cls.data_stream_types:
            size += t.encoded_size()
        return size


class BleDataType(object):
    @classmethod
    def encode(cls, value):
        raise NotImplementedError()

    @classmethod
    def decode(cls, stream):
        """
        :type stream: BleDataStream
        """
        raise NotImplementedError()

    @classmethod
    def encoded_size(cls):
        raise NotImplementedError()


class DoubleNibble(BleDataType):
    @classmethod
    def encode(cls, value):
        # value should be a list of two integers
        v = ((value[0] & 0x0F) << 4) | (value[1] & 0x0F)
        return struct.pack("<B", v)

    @classmethod
    def decode(cls, stream):
        """
        :type stream: BleDataStream
        """
        value = struct.unpack("<B", stream.take(1))[0]
        return [value >> 4 & 0x0F, value & 0x0F]

    @classmethod
    def encoded_size(cls):
        return 1


class UnsignedIntegerBase(BleDataType):
    signed = False
    byte_count = 1
    _bytes_to_struct_formats = {
        1: "b",
        2: "h",
        4: "i",
        8: "q"
    }

    @classmethod
    def _decode_size(cls):
        return 2**(cls.byte_count - 1).bit_length()

    @classmethod
    def _formatter(cls):
        f = cls._bytes_to_struct_formats[cls._decode_size()]
        return "<{}".format(f if cls.signed else f.upper())

    @classmethod
    def encode(cls, value):
        return struct.pack(cls._formatter(), value)[:cls.byte_count]

    @classmethod
    def decode(cls, stream):
        """
        :type stream: BleDataStream
        """
        value_stream = stream.take(cls.byte_count) + b"\x00" * (cls._decode_size()-cls.byte_count)
        value = struct.unpack(cls._formatter(), value_stream)[0]
        return value

    @classmethod
    def encoded_size(cls):
        return cls.byte_count


class SignedIntegerBase(UnsignedIntegerBase):
    signed = True


class Int8(SignedIntegerBase):
    byte_count = 1


class Uint8(UnsignedIntegerBase):
    byte_count = 1


class Int16(SignedIntegerBase):
    byte_count = 2


class Uint16(UnsignedIntegerBase):
    byte_count = 2


class Uint24(UnsignedIntegerBase):
    byte_count = 3


class Uint32(UnsignedIntegerBase):
    byte_count = 4


class Int32(SignedIntegerBase):
    byte_count = 4


class Uint40(UnsignedIntegerBase):
    byte_count = 5


class Uint48(UnsignedIntegerBase):
    byte_count = 6


class Uint56(UnsignedIntegerBase):
    byte_count = 7


class Uint64(UnsignedIntegerBase):
    byte_count = 8


class Int64(SignedIntegerBase):
    byte_count = 8


class String(BleDataType):
    @classmethod
    def encode(cls, value):
        return unicode(value)

    @classmethod
    def decode(cls, stream):
        # TODO: Are strings null-terminated in BLE types?
        return stream.take_all()


class SFloat(BleDataType):
    class ReservedMantissaValues(object):
        POS_INFINITY = 0x07FE
        NEG_INFINITY = 0x0802
        NAN = 0x07FF
        NRES = 0x0800
        RESERVED = 0x0801

        ALL_NAN = [NAN, NRES, RESERVED]

    _mantissa_max = 0x07FD
    _exponent_max = 7
    _exponent_min = -8
    _sfloat_max = 20450000000
    _sfloat_min = -_sfloat_max
    _epsilon = 1e-8
    _precision = 10000

    @classmethod
    def _encode_value(cls, value):
        # Function taken from here: https://github.com/signove/antidote/blob/master/src/util/bytelib.c#L491
        value = float(value)
        sign = 1.0 if value >= 0 else -1.0
        mantissa = abs(value)
        exponent = 0

        # Convert float into the exponent-mantissa pair.  Exponent is base-10
        while mantissa > cls._mantissa_max:
            mantissa /= 10.0
            exponent += 1

            # Secondary check to ensure exponent is in bounds
            if exponent > cls._exponent_max:
                return cls.ReservedMantissaValues.POS_INFINITY if sign > 0 else cls.ReservedMantissaValues.NEG_INFINITY

        while mantissa < 1:
            mantissa *= 10.0
            exponent -= 1

            # Secondary check to ensure exponent is in bounds
            if exponent < cls._exponent_min:
                return 0

        # scale down if number needs more precision
        s_mantissa = round(mantissa * cls._precision)
        r_mantissa = round(mantissa) * cls._precision
        diff = abs(s_mantissa - r_mantissa)

        while diff > 0.5 and exponent > cls._exponent_min and mantissa*10 <= cls._mantissa_max:
            mantissa *= 10.0
            exponent -= 1
            s_mantissa = round(mantissa * cls._precision)
            r_mantissa = round(mantissa) * cls._precision
            diff = abs(s_mantissa - r_mantissa)

        int_mantissa = int(round(sign * mantissa))
        value = ((exponent & 0x000F) << 12) | (int_mantissa & 0x0FFF)
        return value

    @classmethod
    def encode(cls, value):
        if math.isnan(value):
            value = cls.ReservedMantissaValues.NAN
        if value > cls._sfloat_max:
            value = cls.ReservedMantissaValues.POS_INFINITY
        elif value < cls._sfloat_min:
            value = cls.ReservedMantissaValues.NEG_INFINITY
        else:
            value = cls._encode_value(value)

        return struct.pack("<H", value)

    @classmethod
    def decode(cls, stream):
        # Function taken from here: https://github.com/signove/antidote/blob/master/src/util/bytelib.c#L281
        value_int = struct.unpack("<H", stream.take(2))[0]
        mantissa = value_int & 0x0FFF
        exponent = value_int >> 12

        # Get the 2s complement if the value is greater than 8
        if exponent > cls._exponent_max:
            exponent = -(0x10 - exponent)

        if mantissa >= 0x0800:
            mantissa = -((0x0FFF+1) - mantissa)

        if mantissa == cls.ReservedMantissaValues.POS_INFINITY:
            value = float("inf")
        elif mantissa == cls.ReservedMantissaValues.NEG_INFINITY:
            value = float("-inf")
        elif mantissa in cls.ReservedMantissaValues.ALL_NAN:
            value = float("nan")
        else:
            value = mantissa * 10.0**exponent

        return value

    @classmethod
    def encoded_size(cls):
        return 2  # 16-bit


class DateTime(BleCompoundDataType):
    data_stream_types = [Uint16, Uint8, Uint8, Uint8, Uint8, Uint8]

    def __init__(self, dt):
        """
        :type dt: datetime.datetime
        """
        self.dt = dt

    def encode(self):
        return self.encode_values(self.dt.year, self.dt.month, self.dt.day,
                                  self.dt.hour, self.dt.minute, self.dt.second)

    @classmethod
    def decode(cls, stream):
        y, mo, d, h, m, s = super(DateTime, cls).decode(stream)
        return datetime.datetime(y, mo, d, h, m, s)


class DayOfWeek(IntEnum):
    unknown = 0
    monday = 1
    tuesday = 2
    wednesday = 3
    thursday = 4
    friday = 5
    saturday = 6
    sunday = 7


class DayDateTime(BleCompoundDataType):
    data_stream_types = [DateTime, Uint8]

    def __init__(self, dt):
        """
        :type dt: datetime.datetime
        """
        self.dt = dt

    def encode(self):
        stream = DateTime(self.dt).encode()
        # datetime weekdays are 0-6 (mon-sun), convert to 1-7 (mon-sun)
        weekday = self.dt.weekday() + 1
        stream.encode(Uint8, weekday)
        return stream

    @classmethod
    def decode(cls, stream):
        """
        :rtype: datetime.datetime
        """
        dt, day_of_week = super(DayDateTime, cls).decode(stream)
        return dt  # Do we care about the day of week at this point?


class Bitfield(BleCompoundDataType):
    bitfield_width = 8
    bitfield_enum = None

    _width_type_map = {
        8: Uint8,
        16: Uint16,
        24: Uint24,
        32: Uint32,
        40: Uint40,
        48: Uint48,
        56: Uint56,
        64: Uint64,
    }

    def __init__(self):
        assert self.bitfield_width % 8 == 0
        assert type(self.bitfield_enum), IntEnum
        self._mapping = {enum.value: enum.name for enum in self.bitfield_enum}
        assert max(self._mapping.keys()) < self.bitfield_width

    def _iter_bits(self):
        for bit, attr_name in sorted(self._mapping.items()):
            yield bit, attr_name

    def encode(self):
        stream = BleDataStream()
        value = 0
        for bit, attr_name in self._iter_bits():
            bit_value = getattr(self, attr_name)
            if bit_value:
                value |= 1 << bit
        stream.encode(self._encoder_class(), value)
        return stream

    @classmethod
    def _encoder_class(cls):
        return cls._width_type_map[cls.bitfield_width]

    @classmethod
    def decode(cls, stream):
        value = cls._encoder_class().decode(stream)
        return cls.from_integer_value(value)

    @classmethod
    def from_integer_value(cls, value):
        bitfield = cls()
        for bit, attr_name in bitfield._iter_bits():
            if (value & (1 << bit)) > 0:
                setattr(bitfield, attr_name, True)

        return bitfield

    @classmethod
    def byte_count(cls):
        return cls._encoder_class().byte_count

    @classmethod
    def encoded_size(cls):
        return cls.byte_count()

    def __repr__(self):
        set_bit_strs = []

        for bit, attr_name in self._iter_bits():
            if getattr(self, attr_name):
                set_bit_strs.append("{}({})".format(attr_name, bit))
        return "{}({})".format(self.__class__.__name__, ", ".join(set_bit_strs))



import binascii

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes


# Elliptic Curve used for LE Secure connections
_lesc_curve = ec.SECP256R1
_backend = default_backend()


def lesc_pubkey_to_raw(public_key, little_endian=True):
    """
    Converts from a python public key to the raw (x, y) bytes for the nordic
    """
    pk = public_key.public_numbers()
    x = bytearray.fromhex("{:064x}".format(pk.x))
    y = bytearray.fromhex("{:064x}".format(pk.y))

    # Nordic requires keys in little-endian, rotate
    if little_endian:
        pubkey_raw = x[::-1] + y[::-1]
    else:
        pubkey_raw = x + y
    return pubkey_raw


def lesc_privkey_to_raw(private_key, little_endian=True):
    pk = private_key.private_numbers()
    x = bytearray.fromhex("{:032x}".format(pk.private_value))
    if little_endian:
        x = x[::-1]
    return x


def lesc_pubkey_from_raw(raw_key, little_endian=True):
    """
    Converts from raw (x, y) bytes to a public key that can be used for the DH request
    """
    key_len = len(raw_key)
    x_raw = raw_key[:key_len//2]
    y_raw = raw_key[key_len//2:]

    # Nordic transmits keys in little-endian, convert to big-endian
    if little_endian:
        x_raw = x_raw[::-1]
        y_raw = y_raw[::-1]

    x = int(binascii.hexlify(x_raw), 16)
    y = int(binascii.hexlify(y_raw), 16)
    public_numbers = ec.EllipticCurvePublicNumbers(x, y, _lesc_curve())
    return public_numbers.public_key(_backend)


def lesc_privkey_from_raw(raw_priv_key, raw_pub_key, little_endian=True):
    key_len = len(raw_pub_key)
    x_raw = raw_pub_key[:key_len//2]
    y_raw = raw_pub_key[key_len//2:]

    if little_endian:
        x_raw = x_raw[::-1]
        y_raw = y_raw[::-1]
        raw_priv_key = raw_priv_key[::-1]

    x = int(binascii.hexlify(x_raw), 16)
    y = int(binascii.hexlify(y_raw), 16)
    priv = int(binascii.hexlify(raw_priv_key), 16)
    public_numbers = ec.EllipticCurvePublicNumbers(x, y, _lesc_curve())
    priv_numbers = ec.EllipticCurvePrivateNumbers(priv, public_numbers)
    return priv_numbers.private_key(_backend)


def lesc_generate_private_key():
    """
    Generates a new private key that can be used for LESC pairing

    :return: The generated private key
    """
    return ec.generate_private_key(_lesc_curve, _backend)


def lesc_compute_dh_key(private_key, peer_public_key, little_endian=False):
    """
    Computes the DH key for LESC pairing given our private key and the peer's public key

    :param private_key: Our private key
    :param peer_public_key: The peer's public key
    :param little_endian: whether or not to return the shared secret in little endian
    :return: The shared secret
    """
    dh_key = private_key.exchange(ec.ECDH(), peer_public_key)
    if little_endian:
        dh_key = dh_key[::-1]
    return dh_key


def ble_ah(key, p_rand):
    """
    Function for calculating the ah() hash function described in Bluetooth core specification 4.2 section 3.H.2.2.2.

    This is used for resolving private addresses where a private address
    is prand[3] || aes-128(irk, prand[3]) % 2^24

    :param key: the IRK to use, in big endian format
    :param p_rand:
    :return:
    """
    if len(p_rand) != 3:
        raise ValueError("Prand must be a str or bytes of length 3")

    # prepend the prand with 0's to fill up a 16-byte block
    p_rand = b"\x00" * 13 + p_rand

    cipher = Cipher(algorithms.AES(key), modes.ECB(), _backend)
    encryptor = cipher.encryptor()

    # Encrypt and return the last 3 bytes
    encrypted_hash = encryptor.update(p_rand) + encryptor.finalize()
    return encrypted_hash[-3:]


def private_address_resolves(peer_addr, irk):
    """
    Checks if the given peer address can be resolved with the IRK

    Private Resolvable Peer Addresses are in the format
    [4x:xx:xx:yy:yy:yy], where 4x:xx:xx is a random number hashed with the IRK to generate yy:yy:yy
    This function checks if the random number portion hashed with the IRK equals the hashed part of the address

    :param peer_addr: The peer address to check
    :param irk: The identity resolve key to try
    :return: True if it resolves, False if not
    """
    # prand consists of the first 3 MSB bytes of the peer address
    p_rand = bytes(peer_addr.addr[:3])
    # the calculated hash is the last 3 LSB bytes of the peer address
    addr_hash = bytes(peer_addr.addr[3:])
    # IRK is stored in little-endian bytearray, convert to string and reverse
    irk = bytes(irk)[::-1]
    local_hash = ble_ah(irk, p_rand)
    return local_hash == addr_hash


# BLE LESC Debug keys, defined in the Core Bluetooth Specification v4.2 Vol.3, Part H, Section 2.3.5.6.1
# Keys are in big-endian

_LESC_DEBUG_PRIVATE_KEY_RAW = binascii.unhexlify("3f49f6d4a3c55f3874c9b3e3d2103f504aff607beb40b7995899b8a6cd3c1abd")
_LESC_DEBUG_PUBLIC_KEY_X_RAW = binascii.unhexlify("20b003d2f297be2c5e2c83a7e9f9a5b9eff49111acf4fddbcc0301480e359de6")
_LESC_DEBUG_PUBLIC_KEY_Y_RAW = binascii.unhexlify("dc809c49652aeb6d63329abf5a52155c766345c28fed3024741c8ed01589d28b")
_LESC_DEBUG_PUBLIC_KEY_RAW = _LESC_DEBUG_PUBLIC_KEY_X_RAW + _LESC_DEBUG_PUBLIC_KEY_Y_RAW

LESC_DEBUG_PRIVATE_KEY = lesc_privkey_from_raw(_LESC_DEBUG_PRIVATE_KEY_RAW, _LESC_DEBUG_PUBLIC_KEY_RAW, little_endian=False)
LESC_DEBUG_PUBLIC_KEY = LESC_DEBUG_PRIVATE_KEY.public_key()



from __future__ import annotations
import logging
import threading
import enum

from blatann.event_type import EventSource, Event
from blatann.gap import smp
from blatann.gatt import gattc, service_discovery, MTU_SIZE_DEFAULT, MTU_SIZE_MINIMUM
from blatann.nrf import nrf_events
from blatann.nrf.nrf_types.enums import BLE_CONN_HANDLE_INVALID
from blatann.waitables import connection_waitable, event_waitable
from blatann.event_args import *

logger = logging.getLogger(__name__)


class PeerState(enum.Enum):
    DISCONNECTED = 0
    CONNECTING = 1
    CONNECTED = 2


class PeerAddress(nrf_events.BLEGapAddr):
    pass


class ConnectionParameters(nrf_events.BLEGapConnParams):
    def __init__(self, min_conn_interval_ms, max_conn_interval_ms, timeout_ms, slave_latency=0):
        # TODO: Parameter validation
        super(ConnectionParameters, self).__init__(min_conn_interval_ms, max_conn_interval_ms, timeout_ms, slave_latency)


DEFAULT_CONNECTION_PARAMS = ConnectionParameters(15, 30, 4000, 0)
DEFAULT_SECURITY_PARAMS = smp.SecurityParameters(reject_pairing_requests=True)


class Peer(object):
    """
    Object that represents a BLE-connected (or disconnected) peer
    """
    BLE_CONN_HANDLE_INVALID = BLE_CONN_HANDLE_INVALID

    """ Number of bytes that are header/overhead per MTU when sending a notification or indication """
    NOTIFICATION_INDICATION_OVERHEAD_BYTES = 3

    def __init__(self, ble_device, role, connection_params=DEFAULT_CONNECTION_PARAMS,
                 security_params=DEFAULT_SECURITY_PARAMS):
        """
        :type ble_device: blatann.device.BleDevice
        """
        self._ble_device = ble_device
        self._role = role
        self._ideal_connection_params = connection_params
        self._current_connection_params = DEFAULT_CONNECTION_PARAMS
        self.conn_handle = BLE_CONN_HANDLE_INVALID
        self.peer_address = "",
        self.connection_state = PeerState.DISCONNECTED
        self._on_connect = EventSource("On Connect", logger)
        self._on_disconnect = EventSource("On Disconnect", logger)
        self._on_mtu_exchange_complete = EventSource("On MTU Exchange Complete", logger)
        self._on_mtu_size_updated = EventSource("On MTU Size Updated", logger)
        self._mtu_size = MTU_SIZE_DEFAULT
        self._preferred_mtu_size = MTU_SIZE_DEFAULT
        self._negotiated_mtu_size = None

        self._connection_based_driver_event_handlers = {}
        self._connection_handler_lock = threading.Lock()
        self.security = smp.SecurityManager(self._ble_device, self, security_params)

    """
    Properties
    """

    @property
    def connected(self):
        """
        Gets if this peer is currently connected

        :return: True if connected, False if not
        """
        return self.connection_state == PeerState.CONNECTED

    @property
    def bytes_per_notification(self):
        """
        Gets the maximum number of bytes that can be sent in a single notification/indication

        :return: Number of bytes that can be sent in a notification/indication
        """
        return self._mtu_size - self.NOTIFICATION_INDICATION_OVERHEAD_BYTES

    @property
    def is_peripheral(self):
        """
        Gets if this peer is a Peripheral (the local device acting as a central/client)
        """
        return isinstance(self, Peripheral)

    @property
    def is_client(self):
        """
        Gets if this peer is a Client (the local device acting as a peripheral/server)
        """
        return isinstance(self, Client)

    @property
    def is_previously_bonded(self):
        """
        Gets if the peer this security manager is for was bonded in a previous connection
        """
        return self.security.is_previously_bonded

    @property
    def mtu_size(self):
        """
        Gets the current negotiated size of the MTU for the peer

        :return: The current MTU size
        """
        return self._mtu_size

    @property
    def max_mtu_size(self):
        """
        The maximum allowed MTU size. This is set when initially configuring the BLE Device
        """
        return self._ble_device.max_mtu_size

    @property
    def preferred_mtu_size(self):
        """
        Gets the user-set preferred MTU size. Defaults to the Default MTU size (23)
        """
        return self._preferred_mtu_size

    @preferred_mtu_size.setter
    def preferred_mtu_size(self, mtu_size):
        """
        Sets the preferred MTU size to use when a MTU Exchange Request is received
        """
        self._validate_mtu_size(mtu_size)
        self._preferred_mtu_size = mtu_size

    """
    Events
    """

    @property
    def on_connect(self) -> Event[Peer, None]:
        """
        Event generated when the peer connects to the local device

        Event Args: None

        :return: an Event which can have handlers registered to and deregistered from
        """
        return self._on_connect

    @property
    def on_disconnect(self) -> Event[Peer, DisconnectionEventArgs]:
        """
        Event generated when the peer disconnects from the local device

        :return: an Event which can have handlers registered to and deregistered from
        """
        return self._on_disconnect

    @property
    def on_mtu_exchange_complete(self) -> Event[Peer, MtuSizeUpdatedEventArgs]:
        """
        Event generated when an MTU exchange completes with the peer

        :return: an Event which can have handlers registered to and deregistered from
        :rtype: blatann.event_type.Event
        """
        return self._on_mtu_exchange_complete

    @property
    def on_mtu_size_updated(self) -> Event[Peer, MtuSizeUpdatedEventArgs]:
        """
        Event generated when the effective MTU size has been updated on the connection.

        :return: an Event which can have handlers registered to and deregistered from
        :rtype: blatann.event_type.Event
        """
        return self._on_mtu_size_updated

    """
    Public Methods
    """

    def disconnect(self, status_code=nrf_events.BLEHci.remote_user_terminated_connection):
        """
        Disconnects from the peer, giving the optional status code.
        Returns a waitable that will fire when the disconnection is complete

        :param status_code: The HCI Status code to send back to the peer
        :return: A waitable that will fire when the peer is disconnected
        :rtype: connection_waitable.DisconnectionWaitable
        """
        if self.connection_state != PeerState.CONNECTED:
            return
        self._ble_device.ble_driver.ble_gap_disconnect(self.conn_handle, status_code)
        return self._disconnect_waitable

    def set_connection_parameters(self, min_connection_interval_ms, max_connection_interval_ms, connection_timeout_ms,
                                  slave_latency=0):
        """
        Sets the connection parameters for the peer and starts the connection parameter update process

        :param min_connection_interval_ms: The minimum acceptable connection interval, in milliseconds
        :param max_connection_interval_ms: The maximum acceptable connection interval, in milliseconds
        :param connection_timeout_ms: The connection timeout, in milliseconds
        :param slave_latency: The slave latency allowed
        """
        self._ideal_connection_params = ConnectionParameters(min_connection_interval_ms, max_connection_interval_ms,
                                                             connection_timeout_ms, slave_latency)
        if not self.connected:
            return
        # Do stuff to set the connection parameters
        self._ble_device.ble_driver.ble_gap_conn_param_update(self.conn_handle, self._ideal_connection_params)

    def exchange_mtu(self, mtu_size=None):
        """
        Initiates the MTU Exchange sequence with the peer device.

        If the MTU size is not provided the preferred_mtu_size value will be used.
        If an MTU size is provided the preferred_mtu_size will be updated to this

        :param mtu_size: Optional MTU size to use. If provided, it will also updated the preferred MTU size
        :return: A waitable that will fire when the MTU exchange completes
        :rtype: event_waitable.EventWaitable
        """
        # If the MTU size has already been negotiated we need to use the same value
        # as the previous exchange (Vol 3, Part F 3.4.2.2)
        if self._negotiated_mtu_size is None:
            if mtu_size is not None:
                self._validate_mtu_size(mtu_size)
                self._negotiated_mtu_size = mtu_size
            else:
                self._negotiated_mtu_size = self.preferred_mtu_size

        self._ble_device.ble_driver.ble_gattc_exchange_mtu_req(self.conn_handle, self._negotiated_mtu_size)
        return event_waitable.EventWaitable(self._on_mtu_exchange_complete)

    """
    Internal Library Methods
    """

    def peer_connected(self, conn_handle, peer_address, connection_params):
        """
        Internal method called when the peer connects to set up the object
        """
        self.conn_handle = conn_handle
        self.peer_address = peer_address
        self._mtu_size = MTU_SIZE_DEFAULT
        self._negotiated_mtu_size = None
        self._disconnect_waitable = connection_waitable.DisconnectionWaitable(self)
        self.connection_state = PeerState.CONNECTED
        self._current_connection_params = connection_params

        self._ble_device.ble_driver.event_subscribe(self._on_disconnect_event, nrf_events.GapEvtDisconnected)
        self.driver_event_subscribe(self._on_connection_param_update, nrf_events.GapEvtConnParamUpdate, nrf_events.GapEvtConnParamUpdateRequest)
        self.driver_event_subscribe(self._on_mtu_exchange_request, nrf_events.GattsEvtExchangeMtuRequest)
        self.driver_event_subscribe(self._on_mtu_exchange_response, nrf_events.GattcEvtMtuExchangeResponse)
        self.driver_event_subscribe(self._on_data_length_update_request, nrf_events.GapEvtDataLengthUpdateRequest)
        self.driver_event_subscribe(self._on_phy_update_request, nrf_events.GapEvtPhyUpdateRequest)
        self._on_connect.notify(self)

    def _check_driver_event_connection_handle_wrapper(self, func):
        def wrapper(driver, event):
            """
            :param driver:
            :type event: blatann.nrf.nrf_events.BLEEvent
            """
            if self.connected and self.conn_handle == event.conn_handle:
                logger.debug("Got event: {} for peer {}".format(event, self.conn_handle))
                func(driver, event)
        return wrapper

    def driver_event_subscribe(self, handler, *event_types):
        """
        Internal method that subscribes handlers to NRF Driver events directed at this peer.
        Handlers are automatically unsubscribed once the peer disconnects

        :param handler: The handler to subscribe
        :param event_types: The NRF Driver event types to subscribe to
        """
        wrapped_handler = self._check_driver_event_connection_handle_wrapper(handler)
        with self._connection_handler_lock:
            if handler not in self._connection_based_driver_event_handlers:
                self._connection_based_driver_event_handlers[handler] = wrapped_handler
                self._ble_device.ble_driver.event_subscribe(wrapped_handler, *event_types)

    def driver_event_unsubscribe(self, handler, *event_types):
        """
        Internal method that unsubscribes handlers from NRF Driver events

        :param handler: The handler to unsubscribe
        :param event_types: The event types to unsubscribe from
        """
        with self._connection_handler_lock:
            wrapped_handler = self._connection_based_driver_event_handlers.get(handler, None)
            logger.debug("Unsubscribing {} ({})".format(handler, wrapped_handler))
            if wrapped_handler:
                self._ble_device.ble_driver.event_unsubscribe(wrapped_handler, *event_types)
                del self._connection_based_driver_event_handlers[handler]

    """
    Private Methods
    """

    def _on_disconnect_event(self, driver, event):
        """
        :type event: nrf_events.GapEvtDisconnected
        """
        if not self.connected or self.conn_handle != event.conn_handle:
            return
        self.conn_handle = BLE_CONN_HANDLE_INVALID
        self.connection_state = PeerState.DISCONNECTED
        self._on_disconnect.notify(self, DisconnectionEventArgs(event.reason))

        with self._connection_handler_lock:
            for handler in self._connection_based_driver_event_handlers.values():
                self._ble_device.ble_driver.event_unsubscribe_all(handler)
            self._connection_based_driver_event_handlers = {}
        self._ble_device.ble_driver.event_unsubscribe(self._on_disconnect_event)
        self._ble_device.ble_driver.event_unsubscribe(self._on_connection_param_update)

    def _on_connection_param_update(self, driver, event):
        """
        :type event: nrf_events.GapEvtConnParamUpdate
        """
        if not self.connected or self.conn_handle != event.conn_handle:
            return
        if isinstance(event, nrf_events.GapEvtConnParamUpdateRequest):
            logger.debug("[{}] Conn Params updating to {}".format(self.conn_handle, self._ideal_connection_params))
            self._ble_device.ble_driver.ble_gap_conn_param_update(self.conn_handle, self._ideal_connection_params)
        else:
            logger.debug("[{}] Updated to {}".format(self.conn_handle, event.conn_params))
        self._current_connection_params = event.conn_params

    def _validate_mtu_size(self, mtu_size):
        if mtu_size < MTU_SIZE_MINIMUM:
            raise ValueError("Invalid MTU size {}. "
                             "Minimum is {}".format(mtu_size, MTU_SIZE_MINIMUM))
        if mtu_size > self.max_mtu_size:
            raise ValueError("Invalid MTU size {}. "
                             "Maximum configured in the BLE device: {}".format(mtu_size, self._ble_device.max_mtu_size))

    def _resolve_mtu_exchange(self, our_mtu, peer_mtu):
        previous_mtu_size = self._mtu_size
        self._mtu_size = max(min(our_mtu, peer_mtu), MTU_SIZE_MINIMUM)
        logger.debug("[{}] MTU Exchange - Ours: {}, Peers: {}, Effective: {}".format(self.conn_handle,
                                                                                     our_mtu, peer_mtu, self._mtu_size))
        self._on_mtu_size_updated.notify(self, MtuSizeUpdatedEventArgs(previous_mtu_size, self._mtu_size))

        return previous_mtu_size, self._mtu_size

    def _on_mtu_exchange_request(self, driver, event):
        if self._negotiated_mtu_size is None:
            self._negotiated_mtu_size = self.preferred_mtu_size

        self._ble_device.ble_driver.ble_gatts_exchange_mtu_reply(self.conn_handle, self._negotiated_mtu_size)
        self._resolve_mtu_exchange(self._negotiated_mtu_size, event.client_mtu)

    def _on_mtu_exchange_response(self, driver, event):
        previous, current = self._resolve_mtu_exchange(self._negotiated_mtu_size, event.server_mtu)
        self._on_mtu_exchange_complete.notify(self, MtuSizeUpdatedEventArgs(previous, current))

    def _on_data_length_update_request(self, driver, event):
        self._ble_device.ble_driver.ble_gap_data_length_update(self.conn_handle)

    def _on_phy_update_request(self, driver, event):
        self._ble_device.ble_driver.ble_gap_phy_update(self.conn_handle)

    def __nonzero__(self):
        return self.conn_handle != BLE_CONN_HANDLE_INVALID

    def __bool__(self):
        return self.__nonzero__()


class Peripheral(Peer):
    """
    Object which represents a BLE-connected device that is acting as a peripheral/server (local device is client/central)
    """
    def __init__(self, ble_device, peer_address, connection_params=DEFAULT_CONNECTION_PARAMS):
        super(Peripheral, self).__init__(ble_device, nrf_events.BLEGapRoles.central, connection_params)
        self.peer_address = peer_address
        self.connection_state = PeerState.CONNECTING
        self._db = gattc.GattcDatabase(ble_device, self)
        self._discoverer = service_discovery.DatabaseDiscoverer(ble_device, self)

    @property
    def database(self):
        """
        Gets the database on the peripheral.
        NOTE: This is not useful until services are discovered first

        :return: The database instance
        :rtype: gattc.GattcDatabase
        """
        return self._db

    def discover_services(self):
        """
        Starts the database discovery process of the peripheral. This will discover all services, characteristics, and
        descriptors on the remote database.
        Returns an EventWaitable that will fire when the service discovery completes.
        Waitable returns 2  parameters: (Peripheral this, DatabaseDiscoveryCompleteEventArgs event args)

        :return: a Waitable that will fire when service discovery is complete
        :rtype: event_waitable.EventWaitable
        """
        self._discoverer.start()
        return event_waitable.EventWaitable(self._discoverer.on_discovery_complete)


class Client(Peer):
    """
    Object which represents a BLE-connected device that is acting as a client/central (local device is peripheral/server)
    """
    def __init__(self, ble_device, connection_params=DEFAULT_CONNECTION_PARAMS):
        super(Client, self).__init__(ble_device, nrf_events.BLEGapRoles.periph, connection_params)



"""
This example simply demonstrates scanning for peripheral devices
"""
from blatann import BleDevice
from blatann.examples import example_utils

logger = example_utils.setup_logger(level="INFO")


def main(serial_port):
    # Create and open the device
    ble_device = BleDevice(serial_port)
    ble_device.open()

    logger.info("Scanning...")
    # Set scanning for 4 seconds
    ble_device.scanner.set_default_scan_params(timeout_seconds=4)

    # Start scanning and iterate through the reports as they're received
    for report in ble_device.scanner.start_scan().scan_reports:
        if not report.duplicate:
            logger.info(report)

    scan_report = ble_device.scanner.scan_report
    print("\n")
    logger.info("Finished scanning. Scan reports by peer address:")
    # Iterate through all the peers found and print out the reports
    for report in scan_report.advertising_peers_found:
        logger.info(report)

    # Clean up and close the device
    ble_device.close()


if __name__ == '__main__':
    main("COM7")




from setuptools import setup, find_packages
from os import path


VERSION = "v0.3.0-rc.3"

HERE = path.dirname(__file__)
with open(path.join(HERE, "README.md"), "r", encoding="utf-8") as f:
    long_description = f.read().replace("\r", "")


_install_requires = ["pc-ble-driver-py>=0.13", "cryptography", "pytz"]

setup(
    name="blatann",
    version=VERSION.lstrip("v"),  # Remove the leading v, pip doesn't like that
    description="API for controlling nRF52 connectivity devices through pc-ble-driver-py",
    url="https://github.com/ThomasGerstenberg/blatann",
    author="Thomas Gerstenberg",
    email="tgerst6@gmail.com",
    keywords="ble bluetooth nrf52 nordic",
    packages=find_packages(exclude=["test", "test.*"]),
    install_requires=_install_requires,
    python_requires=">=3.7.*",
    long_description_content_type="text/markdown",
    long_description=long_description,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
)



"""
This example demonstrates using Bluetooth SIG's Battery service as a peripheral.
The peripheral adds the service, then updates the battery level percentage periodically.

This can be used in conjunction with the nRF Connect apps to explore the functionality demonstrated
"""
import time
from blatann import BleDevice
from blatann.gap import advertising
from blatann.utils import setup_logger
from blatann.services import battery


logger = setup_logger(level="INFO")


def on_connect(peer, event_args):
    """
    Event callback for when a central device connects to us

    :param peer: The peer that connected to us
    :type peer: blatann.peer.Client
    :param event_args: None
    """
    if peer:
        logger.info("Connected to peer")
    else:
        logger.warning("Connection timed out")


def on_disconnect(peer, event_args):
    """
    Event callback for when the client disconnects from us (or when we disconnect from the client)

    :param peer: The peer that disconnected
    :type peer: blatann.peer.Client
    :param event_args: The event args
    :type event_args: blatann.event_args.DisconnectionEventArgs
    """
    logger.info("Disconnected from peer, reason: {}".format(event_args.reason))


def main(serial_port):
    # Create and open the device
    ble_device = BleDevice(serial_port)
    ble_device.open()

    # Create and add the battery service to the database
    battery_service = battery.add_battery_service(ble_device.database, enable_notifications=True)
    battery_service.set_battery_level(100, False)

    # Register listeners for when the client connects and disconnects
    ble_device.client.on_connect.register(on_connect)
    ble_device.client.on_disconnect.register(on_disconnect)

    # Leaving security and connection parameters as defaults (don't care)

    # Advertise the Battery Service
    adv_data = advertising.AdvertisingData(local_name="Battery Test", flags=0x06,
                                           service_uuid16s=battery.BATTERY_SERVICE_UUID)
    ble_device.advertiser.set_advertise_data(adv_data)

    logger.info("Advertising")
    ble_device.advertiser.start(timeout_sec=0, auto_restart=True)

    battery_level = 100
    battery_level_decrement_steps = 1
    time_between_steps = 10

    # Decrement the battery level until it runs out
    while battery_level >= 0:
        time.sleep(time_between_steps)
        battery_level -= battery_level_decrement_steps
        logger.info("Updating battery level to {}".format(battery_level))
        battery_service.set_battery_level(battery_level)

    logger.info("Done")
    ble_device.close()


if __name__ == '__main__':
    main("COM8")
